import copy
from typing import Dict

from stixorm.module.authorise import authorised_mappings, default_import_type
from stixorm.module.parsing.conversion_decisions import sdo_type_to_tql, sro_type_to_tql, sco__type_to_tql, meta_type_to_tql

from stixorm.module.orm.import_utilities import clean_props, get_embedded_match, split_on_activity_type, \
    add_property_to_typeql, add_relation_to_typeql, val_tql

import logging

from stixorm.module.typedb_lib.factories.auth_factory import get_auth_factory_instance
logger = logging.getLogger(__name__)



marking =["marking-definition--613f2e26-407d-48c7-9eca-b8e91df99dc9",
          "marking-definition--34098fce-860f-48ae-8e50-ebd3cc5e41da",
          "marking-definition--f88d31f6-486f-44da-b317-01333bde0b82",
          "marking-definition--5e57c739-391a-4eb3-b6be-7d15ca92d5ed"]

# ---------------------------------------------------
# 1.0) Helper method to direct the right typeql method to an incoming Stix object
# ---------------------------------------------------


def stix2_to_typeql(stix_object, import_type=default_import_type):
    """
    Initial function to convert Stix into typeql, it adds together the match and insert statements

    Args:
        stix_object (): valid Stix2 object
        import_type (): string, either Stix2 or ATT&CK

    Returns:
        typeql: a string of typeql to match and insert the object int typedb_lib

    """
    match, insert, dep_obj = stix2_to_match_insert(stix_object, import_type)
    typeql = match + insert

    return typeql, dep_obj


def stix2_to_match_insert(stix_object, import_type=default_import_type):
    """
    Initial function to convert Stix into match/insert statments

    Args:
        stix_object (): valid Stix2 object
        import_type (): string, either Stix2 or ATT&CK

    Returns:
        match: a typeql match statement for this object
        insert: a typeql insert statement for this object

    """
    dep_match, dep_insert, indep_ql, core_ql, dep_obj = raw_stix2_to_typeql(stix_object, import_type)
    if dep_match == '':
        match = ''
    else:
        match = 'match ' + dep_match
    if indep_ql == '' and dep_insert == '':
        insert = ''
    else:
        insert = 'insert ' + indep_ql + dep_insert

    return match, insert, dep_obj


def raw_stix2_to_typeql(stix_object,
                        import_type=None) -> [str, str, str, str, {}]:
    """
    Initial function to convert Stix into typeql, it splits the incoming object into different
    channels based on its object type: sdo, sro, sco or meta

    Args:
        stix_object (): valid Stix2 object
        import_type (): string, either Stix2 or ATT&CK

    Returns:
        dep_match: a typeql match statement that depends on other objects
        dep_insert: a typeql insert statement that depends on other objects
        indep_ql: a typeql insert statement with no extenral dependencies
        core_ql: a typeql insert statement that describes the object head, so the independent and dependent parts can be injected seaparately

    """
    if import_type is None:
        import_type = default_import_type

    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    logger.debug(f'stix object type {stix_object["type"]}\n')

    auth_types = copy.deepcopy(auth["types"])
    if stix_object.type in auth_types["sdo"]:
        logger.debug(f' going into sdo ---? {stix_object}')
        dep_match, dep_insert, indep_ql, core_ql, dep_obj = sdo_to_typeql(stix_object, import_type)
    elif stix_object.type in auth_types["sro"]:
        logger.debug(f' going into sro ---> {stix_object}')
        dep_match, dep_insert, indep_ql, core_ql, dep_obj = sro_to_typeql(stix_object, import_type)
    elif stix_object.type in auth_types["sco"]:
        logger.debug(f' going into sco ---> {stix_object}')
        dep_match, dep_insert, indep_ql, core_ql, dep_obj = sco_to_typeql(stix_object, import_type)
    elif stix_object.type == 'marking-definition':
        dep_match, dep_insert, indep_ql, core_ql, dep_obj = marking_definition_to_typeql(stix_object, import_type)
    else:
        logger.error(f'object type not supported: {stix_object.type}, import type {import_type}')
        dep_match, dep_insert, indep_ql, core_ql, dep_obj = '', '', '', '', ''
        dep_list = []

    return dep_match, dep_insert, indep_ql, core_ql, dep_obj


# -------------------------------------------------------------
# 1.1) SDO Object Method to convert a Python object --> typeql string
#                 -   
# -------------------------------------------------------------
def sdo_to_data(sdo, import_type=default_import_type) -> [dict, Dict[str, str], str]:
    """ convert Stix object into a data model for processing

    Args:
        sdo (): the Stix2 SDO object
        import_type (): the type of import to use

    Returns:
        total_props, : a list of all properties
        obj_tql : the dict of the tql properties

    """
    sdo_tql_name = sdo.type
    # - list of property names that have values
    total_props = sdo._inner
    total_props = clean_props(total_props)
    # 1.B) get the specific typeql names for an object into a dictionary
    # b. Instance details
    attack_object = False if not sdo.get("x_mitre_version", False) else True
    sub_technique = False
    if attack_object:
        sub_technique = False if not sdo.get("x_mitre_is_subtechnique", False) else True

    obj_tql, sdo_tql_name, is_list, protocol = sdo_type_to_tql(sdo_tql_name, import_type, attack_object, sub_technique)
    logger.debug(f'\nobject tql {obj_tql}, \nsdo tql name {sdo_tql_name},\n is_list {is_list}')

    return total_props, obj_tql, sdo_tql_name, protocol


def sdo_to_typeql(sdo, import_type=default_import_type) -> [str, str, str, str, dict]:
    """
    Initial function to convert Stix2 SDO object into typeql

    Args:
        sdo (): valid Stix2 object
        import_type (): string, either Stix2 or ATT&CK

    Returns:
        dep_match: a typeql match statement that depends on other objects
        dep_insert: a typeql insert statement that depends on other objects
        indep_ql: a typeql insert statement with no extenral dependencies
        core_ql: a typeql insert statement that describes the object head, so the independent and dependent parts can be injected seaparately

    """
    # 1.A) get configuration parameters
    # - variable for use in typeql statements
    dep_list = []
    # 1.B) get the data model
    total_props, obj_tql, sdo_tql_name, protocol = sdo_to_data(sdo, import_type)
    logger.debug("\n Step 0 I've just gotten through getting data")
    logger.debug(f'\n\n total_props {total_props}\n\nobj_tql {obj_tql}\n\nsdo_tql_name {sdo_tql_name}')
    sdo_var = '$' + sdo_tql_name
    if obj_tql == '':
        return '', '', '', '', {}
    properties, relations = split_on_activity_type(total_props, obj_tql)
    logger.debug("\n----> Step 1 sdo to typeql")

    # 2.) setup the typeql statement for the sdo entity
    sdo_var = '$' + sdo_tql_name
    indep_ql = sdo_var + ' isa ' + sdo_tql_name
    core_ql = sdo_var + ' isa ' + sdo_tql_name + ', has stix-id $stix-id;\n$stix-id ' + val_tql(sdo.id) + ';\n'
    indep_ql_props = dep_match = dep_insert = ''
    logger.debug("----> Step 2 sdo to typeql")
    # 3.) add each of the properties and values of the properties to the typeql statement
    prop_var_list = []
    for prop in properties:
        # split off for properties processing
        indep_ql2, indep_ql_props2, prop_var_list = add_property_to_typeql(prop, obj_tql, sdo, prop_var_list)
        # then add them all together
        indep_ql += indep_ql2
        indep_ql_props += indep_ql_props2
        # add a terminator on the end of the query statement
    indep_ql += ";\n" + indep_ql_props + "\n\n"
    logger.debug("----> Step 3 sdo to typeql")

    # 4.) add each of the relations to the match and insert statements
    for j, rel in enumerate(relations):
        # split off for relation processing
        dep_match2, dep_insert2, dep_list2 = add_relation_to_typeql(rel, sdo, sdo_var, prop_var_list, import_type, j, protocol)
        # then add it back together
        dep_match = dep_match + dep_match2
        dep_insert = dep_insert + dep_insert2
        dep_list = dep_list + dep_list2

    logger.debug("----> Step 4 sdo to typeql")
    dep_obj = {"id": sdo.id, "dep_list": dep_list, "type": sdo.type}
    return dep_match, dep_insert, indep_ql, core_ql, dep_obj


# -------------------------------------------------------
# 1.2) SRO Object Method to convert a Python object --> typeql string
#                 -   
# -----------------------------------------------------
def sro_to_data(sro, import_type=default_import_type) -> [dict, Dict[str, str], str]:
    """ convert Stix object into a data model for processing

        Args:
            sro (): the Stix2 sco object
            import_type (): the type of import to use

        Returns:
            total_props, : a list of all properties
            obj_tql : the dict of the twl proeprties

    """
    # - list of property names that have values, and do not include False values
    total_props = sro._inner
    total_props = clean_props(total_props)

    logger.debug(f'into sro -> {sro}')
    # - work out the type of object
    uses_relation = False
    is_procedure = False
    attack_object = False if not sro.get("x_mitre_version", False) else True
    if attack_object:
        uses_relation = False if not sro.get("relationship_type", False) == "uses" else True
        is_procedure = False if not sro.get("target_ref", False) == "attack-pattern" else True
    obj_tql = {}
    sro_tql_name = sro.type
    sro_sub_rel = "" if not sro.get("relationship_type", False) else sro["relationship_type"]

    obj_tql, sro_tql_name, is_list, protocol = sro_type_to_tql(sro_tql_name, sro_sub_rel, import_type, attack_object, uses_relation, is_procedure)
    logger.debug(f'object tql {obj_tql}, sro tql name {sro_tql_name}')

    return total_props, obj_tql, sro_tql_name, protocol


def sro_to_typeql(sro, import_type=default_import_type) -> [str, str, str, str, dict]:
    """
    Initial function to convert Stix2 SRO object into typeql

    Args:
        sro (): valid Stix2 object
        import_type (): string, either Stix2 or ATT&CK

    Returns:
        dep_match: a typeql match statement that depends on other objects
        dep_insert: a typeql insert statement that depends on other objects
        indep_ql: a typeql insert statement with no extenral dependencies
        core_ql: a typeql insert statement that describes the object head, so the independent and dependent parts can be injected seaparately

    """
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    # 1.) get configuration parameters
    # - variable for use in typeql statements
    dep_list = []
    # - work out the type of object
    obj_type = sro.type
    total_props, obj_tql, sro_tql_name, protocol = sro_to_data(sro, import_type)
    sro_var = '$' + sro_tql_name
    if obj_tql == '':
        return '', '', '', '', {}
    # initialise the typeql insert statement
    dep_match = dep_insert = indep_ql = core_ql = dep_insert_props = ''

    # 2.) setup the match statements first, depending on whether the object is a sighting or a relationship
    # A. If it is a Relationship then find the source and target roles for the relation, and match them in
    if obj_type == 'relationship':
        source_id = sro.source_ref
        dep_list.append(source_id)
        source_var, source_match = get_embedded_match(source_id, 0, protocol, import_type)
        target_id = sro.target_ref
        dep_list.append(target_id)
        target_var, target_match = get_embedded_match(target_id, 0, protocol, import_type)
        dep_match += source_match + target_match
        # 3.)  then setup the typeql statement to insert the specific sro relation, from the dict, with the matches
        for record in auth["reln"]["standard_relations"]:
            if record['stix'] == sro_tql_name:
                dep_insert += '\n' + sro_var
                dep_insert += ' (' + record['source'] + ':' + source_var
                dep_insert += ', ' + record['target'] + ':' + target_var + ')'
                dep_insert += ' isa ' + record['typeql']
                core_ql = sro_var + ' isa ' + sro_tql_name
                core_ql += ', has stix-id $stix-id;\n$stix-id ' + val_tql(sro.id) + ';\n'
                break
                # B. If it is a Sighting then match the object to the sighting
        logger.debug(f'dep_insert -> {dep_insert}')
    elif obj_type == 'sighting':
        sighting_of_id = sro.sighting_of_ref
        dep_list.append(sighting_of_id)
        sighting_of_var, sighting_of_match = get_embedded_match(sighting_of_id, 0, protocol, import_type)
        dep_match += ' \n' + sighting_of_match
        dep_insert += '\n' + sro_var + ' (sighting-of:' + sighting_of_var
        # if there is observed data list, then add it to the match statement
        observed_data_list = sro.get("observed_data_refs")
        if (observed_data_list is not None) and (len(observed_data_list) > 0):
            for i, observed_data_id in enumerate(observed_data_list):
                dep_list.append(observed_data_id)
                observed_data_var, observed_data_match = get_embedded_match(observed_data_id, i, protocol, import_type)
                dep_match += observed_data_match
                dep_insert += ', observed:' + observed_data_var
        # if there is a list of who and where the sighting's occured, then match it in
        where_sighted_list = sro.get("where_sighted_refs")
        if (where_sighted_list is not None) and (len(where_sighted_list) > 0):
            for where_sighted_id in where_sighted_list:
                dep_list.append(where_sighted_id)
                where_sighted_var, where_sighted_match = get_embedded_match(where_sighted_id, 1, protocol, import_type)
                dep_match += where_sighted_match
                dep_insert += ', where-sighted:' + where_sighted_var

        # then finalise the typeql statement for the sro sighting
        dep_insert += ') isa sighting'
        core_ql = sro_var + ' ($role:$any) isa sighting'
        core_ql += ', has stix-id $stix-id;\n$stix-id ' + val_tql(sro.id) + ';\n'
    else:
        logger.error(f'relationship type {obj_type} not supported')
        return ''

    # 4.) next, split total properties into actual properties and nested structures (Relations)
    properties, relations = split_on_activity_type(total_props, obj_tql)

    # 5.) add each of the properties and values of the properties to the typeql statement
    prop_var_list = []
    for prop in properties:
        # split off for properties processing
        dep_insert2, dep_insert_props2, prop_var_list = add_property_to_typeql(prop, obj_tql, sro, prop_var_list)
        # then add them all together
        dep_insert += dep_insert2
        dep_insert_props += dep_insert_props2
        # add a terminator on the end of the query statement
    dep_insert += ";\n" + dep_insert_props + "\n\n"

    # 6.) add each of the relations to the match and insert statements
    for j, rel in enumerate(relations):
        # split off for relation processing
        dep_match2, dep_insert2, dep_list2 = add_relation_to_typeql(rel,
                                                                    sro,
                                                                    sro_var,
                                                                    prop_var_list,
                                                                    import_type,
                                                                    j,
                                                                    protocol)
        # then add it back together
        dep_match = dep_match + dep_match2
        dep_insert = dep_insert + dep_insert2
        dep_list = dep_list + dep_list2

    dep_obj = {"id": sro.id, "dep_list": dep_list, "type": "relation"}
    return dep_match, dep_insert, indep_ql, core_ql, dep_obj


# ---------------------------------------------------
# 1.3) SCO Object Method to convert a Python object --> typeql string
#                 -
# --------------------------------------------------
def sco_to_data(sco, import_type=default_import_type) -> [dict, dict, str]:
    """ convert Stix object into a data model for processing

        Args:
            sco (): the Stix2 sco object
            import_type (): the type of import to use

        Returns:
            total_props, : a list of all properties
            obj_tql : the dict of the twl proeprties

    """
    # - list of property names that have values
    total_props = sco._inner
    total_props = clean_props(total_props)
    # logger.debug(properties)
    # - work out the type of object
    sco_tql_name = sco.type
    # - get the object-specific typeql names, sighting or relationship
    obj_tql, sco_tql_name, is_list, protocol = sco__type_to_tql(sco_tql_name, import_type)

    return total_props, obj_tql, sco_tql_name, protocol


def sco_to_typeql(sco, import_type=default_import_type):
    """
    Initial function to convert Stix2 SCO object into typeql

    Args:
        sco (): valid Stix2 object
        import_type (): string, either Stix2 or ATT&CK

    Returns:
        dep_match: a typeql match statement that depends on other objects
        dep_insert: a typeql insert statement that depends on other objects
        indep_ql: a typeql insert statement with no extenral dependencies
        core_ql: a typeql insert statement that describes the object head, so the independent and dependent parts can be injected seaparately

    """
    # 1.) get configuration parameters
    # - variable for use in typeql statements
    sco_var = '$' + sco.type
    dep_list = []
    # initialise the typeql insert statement
    dep_match = dep_insert = indep_ql = core_ql = dep_insert_props = ''

    # 1.C) Split them into properties and relations
    total_props, obj_tql, sco_tql_name, protocol = sco_to_data(sco, import_type)
    properties, relations = split_on_activity_type(total_props, obj_tql)

    # 2.) setup the typeql statement for the sco entity
    dep_insert = sco_var + ' isa ' + sco.type
    core_ql = sco_var + ' isa ' + sco.type + ', has stix-id $stix-id;\n$stix-id ' + val_tql(sco.id) + ';\n'

    # 3.) add each of the properties and values of the properties to the typeql statement
    # 5.) add each of the properties and values of the properties to the typeql statement
    prop_var_list = []
    for prop in properties:
        # split off for properties processing
        dep_insert2, dep_insert_props2, prop_var_list = add_property_to_typeql(prop, obj_tql, sco, prop_var_list)
        # then add them all together
        dep_insert += dep_insert2
        dep_insert_props += dep_insert_props2
        # add a terminator on the end of the insert statement
    dep_insert += ";\n" + dep_insert_props + "\n\n"

    # 6.) add each of the relations to the match and insert statements
    for j, rel in enumerate(relations):
        # split off for relation processing
        dep_match2, dep_insert2, dep_list2 = add_relation_to_typeql(rel, sco, sco_var, prop_var_list, import_type, j, protocol)
        # then add it back together
        dep_match = dep_match + dep_match2
        dep_insert = dep_insert + dep_insert2
        dep_list = dep_list + dep_list2

    dep_obj = {"id": sco.id, "dep_list": dep_list, "type": sco.type}
    return dep_match, dep_insert, indep_ql, core_ql, dep_obj


# ---------------------------------------------------
# 1.4) Meta Object Method to convert a Python object --> typeql string
#                 -  marking definitions, statement, colour and Mitre copyright
# --------------------------------------------------


def marking_definition_to_typeql(meta, import_type=default_import_type):
    """
    Initial function to convert Stix2 marking object into typeql

    Args:
        meta (): valid Stix2 object
        import_type (): string, either Stix2 or ATT&CK

    Returns:
        dep_match: a typeql match statement that depends on other objects
        dep_insert: a typeql insert statement that depends on other objects
        indep_ql: a typeql insert statement with no extenral dependencies
        core_ql: a typeql insert statement that describes the object head, so the independent and dependent parts can be injected seaparately

    """
    total_props = meta._inner
    total_props = clean_props(total_props)
    dep_list = []
    statement = {}
    dep_match = dep_insert = indep_ql = core_ql = ''
    # 1.A) if one of the existing colours, return an empty string
    if meta.id in marking:
        return dep_match, dep_insert, indep_ql, core_ql, {}
    # 1.B) Test for attack object and handle statement if a statement marking
    attack_object = False if not meta.get("x_mitre_attack_spec_version", False) else True
    if total_props.get("definition", False):
        statement = total_props["definition"]
        total_props.update(statement)

    obj_tql, meta_tql_name, is_list, protocol = meta_type_to_tql(meta.type, import_type, attack_object)

    properties, relations = split_on_activity_type(total_props, obj_tql)

    # 2.) setup the typeql statement for the sdo entity
    meta_var = '$' + meta_tql_name
    indep_ql = meta_var + ' isa ' + meta_tql_name
    core_ql = meta_var + ' isa ' + meta_tql_name + ', has stix-id $stix-id;\n$stix-id ' + val_tql(meta.id) + ';\n'
    indep_ql_props = dep_match = dep_insert = ''
    logger.debug("----> Step 2 meta to typeql")
    # 3.) add each of the properties and values of the properties to the typeql statement
    prop_var_list = []
    for prop in properties:
        # split off for properties processing
        indep_ql2, indep_ql_props2, prop_var_list = add_property_to_typeql(prop, obj_tql, meta, prop_var_list)
        # then add them all together
        indep_ql += indep_ql2
        indep_ql_props += indep_ql_props2
        # add a terminator on the end of the query statement
    indep_ql += ";\n" + indep_ql_props + "\n\n"
    logger.debug("----> Step 3 sdo to typeql")

    # 4.) add each of the relations to the match and insert statements
    for j, rel in enumerate(relations):
        # split off for relation processing
        dep_match2, dep_insert2, dep_list2 = add_relation_to_typeql(rel, meta, meta_var, prop_var_list, import_type, j,
                                                                    protocol)
        # then add it back together
        dep_match = dep_match + dep_match2
        dep_insert = dep_insert + dep_insert2
        dep_list = dep_list + dep_list2

    logger.debug("----> Step 4 sdo to typeql")
    dep_obj = {"id": meta.id, "dep_list": dep_list, "type": meta.type}
    return dep_match, dep_insert, indep_ql, core_ql, dep_obj

    # # if the marking is a colour, match it in, else it is a statement type
    # if stix_object.definition_type == "statement":
    #     if attack_object:
    #         indep_ql = '\n $marking isa attack-marking'
    #         indep_ql += ',\n has x-mitre-attack-spec-version ' + val_tql(stix_object.x_mitre_attack_spec_version)
    #         loc_list = stix_object.x_mitre_domains
    #         for dom in loc_list:
    #             indep_ql += ',\n has x-mitre-domains ' + val_tql(dom)
    #         core_ql = '$marking isa attack-marking'
    #     else:
    #         indep_ql = '\n $marking isa statement-marking'
    #         core_ql = '$marking isa statement-marking'
    #     indep_ql += ',\n has statement ' + val_tql(stix_object.definition.statement)
    #     indep_ql += ',\n has stix-type "marking-definition"'
    #     indep_ql += ',\n has stix-id ' + val_tql(stix_object.id)
    #     indep_ql += ',\n has created ' + val_tql(stix_object.created)
    #     indep_ql += ',\n has spec-version ' + val_tql(stix_object.spec_version)
    #     indep_ql += ';\n'
    #     core_ql += ', has stix-id $stix-id;\n$stix-id ' + val_tql(stix_object.id)
    #     core_ql += ';'
    #
    # dep_obj = {"id": stix_object.id, "dep_list": dep_list, "type": "marking"}
    # return dep_match, dep_insert, indep_ql, core_ql, dep_obj



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Loader module for all os_threat data model
"""

__author__ = "Brett Forbes"
__credits__ = ["Paolo Di Prodi"]
__license__ = "GPL"
__version__ = "0.1.0"
__maintainer__ = "Paolo Di Prodi"
__email__ = "paolo@priam.ai"
__status__ = "Production"

import json
from glob import glob

import pathlib
from loguru import logger
import os
from pathlib import Path

from stixorm.module.definitions.os_threat.classes import (
    Feeds, Feed, ThreatSubObject, StateChangeObject, EventCoreExt,
    Event, EntityCountObject, ImpactCoreExt, Availability, Confidentiality,
    External, Integrity, Monetary, Physical, Traceability, Impact,
    IncidentScoreObject, IncidentCoreExt, TaskCoreExt, Task,
    EvidenceCoreExt, Evidence
)





name = "os_threat"
class_model = {}
class_model["sdo"] = {
    "Feeds": Feeds,
    "Feed": Feed,
    "Evidence": Evidence,
    "Event": Event,
    "Impact": Impact,
    "Task": Task
}
class_model["sco"] = {}
class_model["sro"] = {}
class_model["sub"] = {
    "ThreatSubObject" : ThreatSubObject,
    "StateChangeObject": StateChangeObject,
    "EventCoreExt": EventCoreExt,
    "EntityCountObject": EntityCountObject,
    "ImpactCoreExt": ImpactCoreExt,
    "Availability": Availability,
    "Confidentiality": Confidentiality,
    "External": External,
    "Integrity": Integrity,
    "Monetary": Monetary,
    "Physical": Physical,
    "Traceability": Traceability,
    "IncidentScoreObject": IncidentScoreObject,
    "IncidentCoreExt": IncidentCoreExt,
    "TaskCoreExt": TaskCoreExt,
    "EvidenceCoreExt": EvidenceCoreExt
}
class_model["meta"] = {}


import traceback


import logging
logger = logging.getLogger(__name__)


def log_delete_layers(result):
    try:
        if result is None:
            logger.debug("Successfully deleted layer")
        else:
            logger.debug("Failed to delete layers")
            logger.exception("\n".join(traceback.format_exception(result.failure())))
    except Exception as e:
        logger.error(e)

def log_delete_layer(result, layer):
    try:
        if result is None:
            logger.debug("Failed to delete layer")
            logger.debug(layer)
    except Exception as e:
        logger.error(e)

def log_add_layer(result, layer):
    try:
        if result is None:
            logger.debug("Failed to add layer")
            logger.debug(layer)
    except Exception as e:
        logger.error(e)

def log_insert_query(result, layer):
    try:
        if result is None:
            logger.debug("Failed to creare insert query")
            logger.debug(layer)
    except Exception as e:
        logger.error(e)

def log_delete_instruction_update_layer(result):
    try:
        if result is None:
            logger.debug("Failed to delete layer")
    except Exception as e:
        logger.error(e)

def log_add_instruction_update_layer(result):
    try:
        if result is None:
            logger.debug("Failed to update layer")
    except Exception as e:
        logger.error(e)


def log_delete_instruction_outcome(result):
    try:
        logger.debug("=========================== delete typeql instruction ====================================")
        if result is None:
            logger.debug("Successfully added instruction")

        else:
            logger.debug("Failed to add the instruction")
            logger.exception("\n".join(traceback.format_exception(result.failure())))
        logger.debug("============================================================================================")

    except Exception as e:
        logger.error(e)

def log_delete_instruction(dep_match,
                           dep_insert,
                           indep_ql,
                           dep_obj,
                           del_match,
                           del_tql):
    try:
        logger.debug(' ---------------------------Delete Object----------------------')
        logger.debug(f'dep_match -> {dep_match}\n dep_insert -> {dep_insert}')
        logger.debug(f'indep_ql -> {indep_ql}\n dep_obj -> {dep_obj}')
        logger.debug("=========================== delete typeql below ====================================")
        logger.debug(f'del_match -> {del_match}\n del_tql -> {del_tql}')
    except Exception as e:
        logger.error(e)



import re
from typing import List
import copy

from stixorm.module.authorise import authorised_mappings

from stixorm.module.orm.import_objects import sdo_to_data, sro_to_data, sco_to_data
from stixorm.module.orm.import_utilities import split_on_activity_type, val_tql

import logging

from stixorm.module.typedb_lib.factories.auth_factory import get_auth_factory_instance

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
##############################################################
#  1.) Methods to Delete any Stix Objects
############################################################

#---------------------------------------------------
# 1.0) Helper method to direct the right typeql method to an incoming Stix object
#---------------------------------------------------


def delete_stix_object(stix_object,
                       dep_match: str,
                       dep_insert: str,
                       indep_ql: str,
                       core_ql: str,
                       import_type) -> [str, str]:
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    if stix_object.type in auth["types"]["sdo"]:
        total_props, obj_tql, sdo_tql_name, protocol = sdo_to_data(stix_object, import_type)
        var_name: List[str] = get_obj_var(indep_ql)
        del_match, del_tql = delete_object(stix_object, core_ql, total_props, obj_tql, var_name, sdo_tql_name, import_type)
    elif stix_object.type in auth["types"]["sro"]:
        total_props, obj_tql, sro_tql_name, protocol = sro_to_data(stix_object, import_type)
        var_name: List[str] = get_obj_var(dep_insert)
        del_match, del_tql = delete_object(stix_object, core_ql, total_props, obj_tql, var_name, sro_tql_name, import_type)
    elif stix_object.type in auth["types"]["sco"]:
        total_props, obj_tql, sro_tql_name, protocol = sco_to_data(stix_object, import_type)
        var_name: List[str] = get_obj_var(core_ql)
        # Need to change this line to suit scenarios where object name is not type name (e.g. future)
        del_match, del_tql = delete_object(stix_object, core_ql, total_props, obj_tql, var_name, sro_tql_name, import_type)
    elif stix_object.type == 'marking-definition':
        del_match, del_tql = delete_marking(stix_object, dep_match, dep_insert, indep_ql, core_ql, import_type)
    else:
        logger.error(f'object type not supported in delete stix object: {stix_object.type}')
        logger.error(f' import type {dep_match + dep_insert + indep_ql + core_ql}')
        del_match = del_tql = ""

    return del_match, del_tql


def delete_object(stix_object,
                  core_ql: str,
                  total_props,
                  obj_tql,
                  var_name: List[str],
                  tql_name: str,
                  import_type) -> [str, str]:
    # 1.B) get the data model
    properties, relations = split_on_activity_type(total_props, obj_tql)
    # 2.0) MAtch in the object, the id and all attributes not owned by another object
    del_match = 'match \n' + core_ql
    del_match += var_name[0] + ' has $a;\n'
    del_match += 'not { ' + var_name[0] + ' isa thing; $p2 isa thing, has $a; not {'
    del_match += var_name[0] + ' is $p2;}; };\n\n'
    del_tql = 'delete \n'
    # 3.0) Now setup the match and delete statements for the local relations or  sub objects
    for i, reln in enumerate(relations):
        del_match2, del_tql2 = delete_sub_reln(reln, stix_object, var_name[0], i, import_type)
        del_match += del_match2
        del_tql += del_tql2

    # 4.0) Now setup the delete and match for the actual attributes, stix-id and object in reverse
    del_tql += '\n'
    del_tql += var_name[0] + ' isa ' + tql_name + ';\n'
    del_tql += '$a isa attribute;\n'
    del_tql += '$stix-id isa stix-id;\n'

    return del_match, del_tql


def delete_marking(stix_object, dep_match, dep_insert, indep_ql, core_ql, import_type) -> [str, str]:
    del_match = del_tql = ''
    if stix_object.definition_type == "statement":
        del_match = 'match \n$marking isa statement-marking'
        del_match += ',\n has stix-id $stix-id; $stix-id ' + val_tql(stix_object.id)
        del_match += ';\n$marking has $a;\n'
        del_match += 'not { $marking isa thing; $p2 isa thing, has $a; not {'
        del_match += '$marking is $p2;}; };\n\n'
        del_tql = 'delete \n'
        del_tql += '$marking isa statement-marking;\n'
        del_tql += '$stix-id isa stix-id;\n'
        del_tql += '$a isa attribute;\n'

    return del_match, del_tql


def get_obj_var(core_ql) -> List[str]:
    m = re.findall(r"(\$[a-z\-0-9]+)\s+(.+)?isa\s+", core_ql,re.MULTILINE)
    output = []
    #logger.debug(f'local 1 -> {m}')
    if m:
        #logger.debug(f'local -> {m}')
        for found in m:
            output.append(found[0])
            #logger.debug(f'Found first {found[0]}')

    return output


def get_tql_name(dep_insert):
    m = re.findall(r"([a-z\-]+),$",dep_insert,re.MULTILINE)
    output = []
    #logger.debug(f'local 1 -> {m}')
    if m:
        for found in m:
            output.append(found)
            #logger.debug(f'Found name {found}')

    return output


def delete_sub_reln(rel, obj, obj_var, i, import_type):
    """
        Top level function to delete one of the sub objects from the stix object
    Args:
        rel (): the relation object to delete
        obj (): the stix object to delete it from
        obj_var (): the typeql variable string

    Returns:
        match: the typeql match string
        delete: the typeql delete string
    """
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    if rel == "granular_markings":
        match, delete = del_granular_markings(obj_var)

    # hashes type
    elif (rel == "hashes"
          or rel == "file_header_hashes"):
        match, delete = del_hashes(rel, obj[rel], obj_var, i)

    # insert key value store
    elif rel in auth["reln_name"]["key_value_relations"]:
        match, delete = del_key_value_store(rel, obj[rel], obj_var, i, import_type)

    # insert list of object relation
    elif rel in auth["reln_name"]["list_of_objects"]:
        match, delete = del_list_of_object(rel, obj[rel], obj_var, i, import_type)

    # insert embedded relations based on stix-id
    elif rel in auth["reln_name"]["embedded_relations"]:
        match, delete = del_embedded_relation(rel, obj[rel], obj_var, i, import_type)

    # insert plain sub-object with relation
    elif (rel == "x509_v3_extensions"
          or rel == "optional_header"):
        match, delete = del_load_object(rel, obj[rel], obj_var, i, import_type)

    # insert  SCO Extensions here, a possible dict of sub-objects
    elif rel in auth["reln_name"]["extension_relations"]:
        match, delete = del_extensions(rel, obj[rel], obj_var, i, import_type)

    # ignore the following relations as they are already processed, for Relationships, Sightings and Extensions
    elif (rel == "sighting_of_ref"
          or rel == "observed_data_refs"
          or rel == "where_sighted_refs"
          or rel == "source_ref"
          or rel == "target_ref"
          or rel == "definition"
          or rel == "definition_type"):
        match = delete = ''

    else:
        logger.error(f'relation type not known, rel -> {rel}')
        match = delete = ""

    return match, delete


def del_granular_markings(obj_var):
    match = '$granular (object:' + obj_var + ') isa granular-marking;\n'
    delete = '$granular isa granular-marking;\n'

    return match, delete


def del_hashes(rel_name, rel_object, obj_var, i):
    match = '$hash isa hash, has hash-value $h;\n'
    match += '$hash_rel (owner:'+obj_var
    match += ', pointed-to:$hash) isa hashes;\n'
    delete = '$hash_rel isa hashes;\n'
    delete += '$hash isa hash;\n'
    return match, delete


def del_key_value_store(rel_name, rel_object, obj_var, i, import_type):
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    for config in auth["reln"]["key_value_relations"]:
        if config["name"] == rel_name:
            rel_typeql = config["typeql"]
            role_owner = config["owner"]
            role_pointed = config["pointed_to"]
            d_key = config["key"]
            d_value = config["value"]
            break

    key_var = '$' + d_key + str(i)
    val_var = ' $' + d_value + str(i)
    match = key_var + ' isa ' + d_key + ';\n'
    match += key_var + ' has ' + d_value + val_var + ';\n'
    match += '$' + rel_typeql + ' (' + role_owner + ':' + obj_var
    match += ', ' + role_pointed + ':' + key_var
    match += ') isa ' + rel_typeql + ';\n\n'
    delete = '\n'
    delete += '$' + rel_typeql + ' isa ' + rel_typeql + ';\n'
    #delete += val_var + ' isa ' + d_value + ';\n'
    #delete += key_var + ' isa ' + d_key + ';\n'

    return match, delete


def del_list_of_object(rel_name, prop_value_list, parent_var, i, import_type):
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    logger.debug(f'rl name -> {rel_name}')
    for config in auth["reln"]["list_of_objects"]:
        if config["name"] == rel_name:
            rel_typeql = config["typeql"]
            role_owner = config["owner"]
            role_pointed = config["pointed_to"]
            typeql_obj = config["object"]
            obj_props_tql = copy.deepcopy(auth["sub_objects"][typeql_obj])
            break
    lod_list = []
    match = delete = ''
    for j, dict_instance in enumerate(prop_value_list):
        lod_var = '$' + typeql_obj + str(j)
        lod_list.append(lod_var)
        match += lod_var + ' isa ' + typeql_obj + ';\n'
        for key in dict_instance:
            typeql_prop = obj_props_tql[key]
            if typeql_prop == '':
                # split off for relation processing
                match2, delete2 = delete_sub_reln(key, dict_instance, lod_var, i + j, import_type)
                # then add it back together
                match = match + match2
                delete = delete + "\n" + delete2 + "\n"
        delete += lod_var + ' isa ' + typeql_obj + ';\n'

    loc_var = '$' + typeql_obj + str(i)
    #match += loc_var + ' isa ' + typeql_obj + ';\n'
    #match += loc_var + ' has $b' + str(i) + ';\n'
    #match += 'not { ' + loc_var + ' isa thing; $lob' + str(i) + ' isa thing, has $b' + str(i) + '; not {'
    #match += loc_var + ' is $lob' + str(i) + ';}; };\n\n'

    match += '\n $' + rel_typeql + str(i) + ' (' + role_owner + ':' + parent_var
    for lod_var in lod_list:
        match += ', ' + role_pointed + ':' + lod_var
    match += ') isa ' + rel_typeql + ';\n'

    #delete += loc_var + ' isa ' + typeql_obj + ';\n'
    #delete += '$b' + str(i) + ' isa attribute; \n'
    delete += '$' + rel_typeql + str(i) + ' isa ' + rel_typeql + ';\n'

    return match, delete


def del_embedded_relation(rel_name, rel_object, obj_var, i, import_type):
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    for ex in auth["reln"]["embedded_relations"]:
      if ex["rel"] == rel_name:
        owner = ex["owner"]
        relation = ex["typeql"]
        break
    loc_var = '$' + relation + str(i)
    match = loc_var + ' (' + owner + ':' + obj_var + ') isa ' + relation + ';\n'
    delete = loc_var + ' isa ' + relation + ';\n'

    return match, delete


def del_load_object(prop_name, prop_dict, parent_var, i, import_type):
    # as long as it is predefined, history the object
    #logger.debug('------------------- history object ------------------------------')
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    logger.debug(f'prop dict {prop_dict}')
    for prop_type in auth["reln"]["extension_relations"]:
        if prop_name == prop_type["stix"]:
            tot_prop_list = [tot for tot in prop_dict.keys()]
            obj_name = prop_type["object"]
            obj_tql = copy.deepcopy(auth["sub_objects"][obj_name])
            obj_var = '$' + obj_name
            reln = prop_type["relation"]
            rel_var = '$' + reln
            rel_owner = prop_type["owner"]
            rel_pointed_to = prop_type["pointed-to"]

            match = obj_var + ' isa ' + prop_type["object"] + ';\n'
            #match += obj_var + ' has $d' + str(i) + ';\n'
            #match += 'not { ' + obj_var + ' isa thing; $lobj' + str(i) + ' isa thing, has $d' + str(i) + '; not {'
            #match += obj_var + ' is $lobj' + str(i) + ';}; };\n\n'

            match += rel_var + ' (' + rel_owner + ':' + parent_var
            match += ', ' + rel_pointed_to + ':' + obj_var + ')'
            match += ' isa ' + reln + ';\n'

            # Split them into properties and relations
            properties, relations = split_on_activity_type(prop_dict, obj_tql)
            delete = ''

            # add each of the relations to the match and insert statements
            for rel in relations:
                # split off for relation processing
                match2, delete2= delete_sub_reln(rel, prop_dict, obj_var, i+1)
                # then add it back together
                match = match + match2
                delete = delete + "\n" + delete2

            # finally, connect the local object to the parent object
            #delete = '$d' + str(i) + ' isa attribute;\n'
            delete += rel_var + ' isa ' + reln + ';\n'
            delete += obj_var + ' isa ' + prop_type["object"] + ';\n'

            break

    return match, delete


def del_extensions(prop_name, prop_dict, parent_var, i, import_type):
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    match = ''
    delete = ''
    # for each key in the dict (extension type)
    # logger.debug('--------------------- extensions ----------------------------')
    for ext_type in prop_dict:
        for ext_type_ql in auth["reln"]["extension_relations"]:
            if ext_type == ext_type_ql["stix"]:
                match2, delete2 = del_load_object(ext_type, prop_dict[ext_type], parent_var, i)
                match = match + match2
                delete = delete + delete2
                break
    return match, delete


def add_delete_layers(layers, dep_obj, indexes, missing):
    logger.debug("################################### enter add_layers ###############################################")
    logger.debug(f'\nlayers -> {layers}')
    logger.debug(f'\ndep_obj -> {dep_obj}')
    logger.debug(f'indexes -> {indexes}')
    logger.debug(f'missing -> {missing}')
    logger.debug("-------------------------------  ------------------------------------------------")
    # Stage 1 - Initialise Variables
    # 1. Setup key variables
    loc_id = dep_obj['id']
    loc_list = dep_obj["dep_list"]
    locset = set(loc_list)
    mset = set(missing)
    iset = set(indexes)
    intset = locset.intersection(iset)
    diffset = locset.difference(intset)
    # Stage 2 - Analyse Choices
    # 2. check whether the object has dependencies in its dep_list
    if not loc_list:
        dep_list_items = False
        logger.debug('### There are no dependencies')
    else:
        dep_list_items = True
        logger.debug('### There are some dependencies')
    # 3. check whether the object id is in the missing list
    if loc_id in mset:
        id_in_missing = True
        logger.debug('### Current is the missing dependency of an existing record')
    else:
        id_in_missing = False
        logger.debug('### Current is not a missing dependency of an existing record')
    # 4. check whether any id in the object dependency list is not already loaded
    if not diffset:
        dep_id_not_loaded = False
        logger.debug('### No dependencies need to be added to missing')
    else:
        dep_id_not_loaded = True
        mset = mset | diffset
        logger.debug('### Missing - Some dependencies need to be added to missing')

    # Stage 3 - Execute Choices
    # 5. No dependencies and not in missing, append only
    if not dep_list_items and not id_in_missing:
        logger.debug('### Append object and return')
        layers.append(dep_obj)
        indexes.append(loc_id)
        logger.debug(f'layers -> {layers}')
        logger.debug(f'indexes -> {indexes}')
        logger.debug(f'mset -> {mset}')
        logger.debug("################################## end of  add_layers ####################################################")
        return layers, indexes, list(mset)
    # 6. There are no dependencies but id is in missing, delete from missing,follow the tree and reorder
    if not dep_list_items and id_in_missing:
        logger.debug('### delete from missing,follow the tree and reorder')
        mset.remove(loc_id)
        tree = follow_the_tree(layers, dep_obj)
        logger.debug(f' tree -> {tree}')
        layers, indexes = reorder(layers, indexes, tree, dep_obj)
        logger.debug(f'layers -> {layers}')
        logger.debug(f'indexes -> {indexes}')
        logger.debug(f'mset -> {mset}')
        logger.debug("################################## end of  add_layers ####################################################")
        return layers, indexes, list(mset)
    # 7 There are dependencies, object is not in missing, insert at front
    if dep_list_items and not id_in_missing:
        logger.debug('### Add current to the front of the record')
        layers.insert(0, dep_obj)
        indexes.insert(0, loc_id)
        logger.debug(f'layers -> {layers}')
        logger.debug(f'indexes -> {indexes}')
        logger.debug(f'mset -> {mset}')
        logger.debug("################################## end of  add_layers ####################################################")
        return layers, indexes, list(mset)
    # 8 There are dependencies, object is in missing , delete from missing,follow the tree and reorder
    if dep_list_items and id_in_missing:
        logger.debug('### delete from missing,follow the tree and reorder')
        mset.remove(loc_id)
        tree = follow_the_tree(layers, dep_obj)
        logger.debug(f' tree -> {tree}')
        layers, indexes = reorder(layers, indexes, tree, dep_obj)
        logger.debug(f'layers -> {layers}')
        logger.debug(f'indexes -> {indexes}')
        logger.debug(f'mset -> {mset}')
        logger.debug("################################## end of  add_layers ####################################################")
        return layers, indexes, list(mset)

    logger.debug("theres a massive problem")
    return layers, indexes, list(mset)


def reorder(layers, indexes, tree, dep_obj):
    front_layers = []
    front_indexes = []
    tree = list(set(tree))
    logger.debug("%%%%%%%%%%%%%%% reorder 1 %%%%%%%%%%%%%%%%%%")
    logger.debug(f'\n orig indexes -> {indexes}')
    logger.debug(f"\n orig layers, {layers}")
    logger.debug(f"\n dep_obj , {dep_obj}")
    logger.debug("%%%%%%%%%%%%%%% reorder 2 %%%%%%%%%%dep_obj%%%%%%%%")
    # 1. Copy elements from layers and indexes so they are in the order we want them
    for t in reversed(tree):
        front_layers.append(layers[t])
        front_indexes.append(indexes[t])
    # 2. Now order the tree in reverse numeric order, biggest first
    tree.sort(reverse=True)
    # 3. Now delete the elements from layers and indexes
    for t in tree:
        layers.pop(t)
        indexes.pop(t)
    # 4. Add the dep_obj to the new lists
    front_indexes.append(dep_obj['id'])
    front_layers.append(dep_obj)
    # 5. Assemble the final lists
    logger.debug(f'\nfront_indexes -> {front_indexes}')
    logger.debug(f"\nfront_layers, {front_layers}")
    logger.debug(f"\n old layers, {layers}")
    logger.debug(f'\nold indexes -> {indexes}')
    logger.debug("-------------------------------------------------------------------------------------")
    layers = front_layers + layers
    indexes = front_indexes + indexes
    #logger.debug(f"\nlayers, {layers}")
    #logger.debug(f'\nindexes -> {indexes}')
    logger.debug("%%%%%%%%%%%%%%% end reorder %%%%%%%%%%%%%%%%%%")
    return layers, indexes


def follow_the_tree(layers, dep_obj):
    tree = []
    loc_id = dep_obj['id']
    loc_ids = [loc_id]
    found = True
    while found:
        found, ret_indexes, ret_ids = find_id(loc_ids, layers)
        logger.debug(f'### Following Tree, found {found},index {ret_indexes}, loc_id {ret_ids}')
        if found:
            tree = tree + ret_indexes
            loc_ids = ret_ids
    return tree


def find_id(loc_ids, layers) -> [bool, List[int], List[str]]:
    found = False
    indexes = []
    ids = []
    for loc_id in loc_ids:
        for index, lay in enumerate(layers):
            if loc_id in lay['dep_list']:
                found = True
                indexes.append(index)
                ids.append(lay['id'])

    return found, indexes, ids






from typing import Dict, List
import copy

from stixorm.module.authorise import  import_type_factory

import logging

from stixorm.module.typedb_lib.factories.auth_factory import get_auth_factory_instance
from stixorm.module.typedb_lib.factories.definition_factory import get_definition_factory_instance
from stixorm.module.typedb_lib.factories.import_type_factory import ImportType
from stixorm.module.typedb_lib.model.definitions import DefinitionName

logger = logging.getLogger(__name__)
default_import_type = import_type_factory.get_default_import()


attack_model = get_definition_factory_instance().lookup_definition(DefinitionName.ATTACK)
stix_model = get_definition_factory_instance().lookup_definition(DefinitionName.STIX_21)
os_threat_model = get_definition_factory_instance().lookup_definition(DefinitionName.OS_THREAT)


def sdo_type_to_tql(sdo_type: str,
                    import_type:ImportType=default_import_type,
                    attack_object=False, subtechnique=False) -> [dict, str, dict, str]:
    """ convert Stix object into a data model for processing

    Args:
        sdo_type (): the Stix2 SDO type
        import_type (): the type of import to use

    Returns:
        obj_tql : the dict of the tql properties
        tql_name : the typeql name of the object
        is_list: a list of all the porperties that are lists

    """
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    obj_tql = {}
    is_list =[]
    protocol = ""
    tql_name = sdo_type
    logger.debug("in sdo decisions")
    logger.debug(f'obj tql {obj_tql}')
    logger.debug(f"variables, stix {auth['STIX21']}, attack is {auth['ATT&CK']}")
    # If import_type is deaful None, then assign to default)

    # 1. get the specific typeql names for an object into a dictionary
    #logger.debug(f'through to decisions, attack is {attack_object}, sub technqiue {subtechnique}')
    #logger.debug(f'auth stix {auth["STIX21"]}, attack auth {auth["ATT&CK"]}')


    if auth['STIX21'] and not auth["ATT&CK"]:
        if stix_model.contains_data(sdo_type):
            # dispatch specific stix properties plus later on, generic sdo properties
            protocol = "stix21"
            obj_tql = copy.deepcopy(stix_model.get_data(sdo_type))
        elif os_threat_model.contains_data(sdo_type):
            # dispatch specific stix properties plus later on, generic sdo properties
            protocol = "os-threat"
            obj_tql = copy.deepcopy(os_threat_model.get_data(sdo_type))
        else:
            logger.error(f'obj_type type {sdo_type} not supported')
            return {}, "", {}, ""
    # - mitre attack_setting import
    elif auth['STIX21'] and auth["ATT&CK"]:
        if attack_object:
            logger.debug("I'm processing an attack decision")
            is_list.extend(auth["is_lists"]["sdo"]["attack"])
            protocol = "attack"
            attack_type = ''
            obj_tql = copy.deepcopy(attack_model.get_base("attack_base"))
            # Convert from stix-type to attack-tql-entity
            for model in attack_model.get_mapping("object_conversion"):
                logger.debug(f'chacking models, type is {model["type"]}')
                if model["type"] == sdo_type:
                    attack_type = model["typeql"]
                    logger.debug(f'attack type is {attack_type}')
                    if attack_type == "technique" and subtechnique:
                        attack_type = "sub-technique"
                    obj_tql.update(attack_model.get_data(attack_type))
                    logger.debug("updated")
                    break
            # Else log an error
            if not attack_type:
                logger.error(f'obj_type type {sdo_type} not in attack type conversion dict, type_to_tql_name')
                return {}, "", {}, ""
            else:
                tql_name = attack_type

        else:
            # its a Stix object, not an AT&CK one
            if stix_model.contains_data(sdo_type):
                # dispatch specific stix properties plus mitre properties plus generic sdo properties
                protocol = "stix21"
                obj_tql = copy.deepcopy(stix_model.get_data(sdo_type))
            elif os_threat_model.contains_data(sdo_type):
                # dispatch specific stix properties plus later on, generic sdo properties
                protocol = "os-threat"
                obj_tql = copy.deepcopy(os_threat_model.get_data(sdo_type))
            else:
                logger.error(f'obj_type type {sdo_type} not in stix_models["dispatch_stix"] or dispatch mitre')
                return {}, "", {}, ""

    else:
        logger.error(f'import type {import_type} not supported')
        return {}, "", {}, ""

    # 1.C) Add the standard object properties to the specific ones, and split them into properties and relations
    logger.debug("about to update stuff")
    logger.debug(f'tql nme {tql_name}, sdo-type {sdo_type}')
    obj_tql.update(stix_model.get_base("base_sdo"))
    is_list.extend(auth["is_lists"]["sdo"][tql_name])
    is_list.extend(auth["is_lists"]["sdo"]["sdo"])
    logger.debug("about to return from decisions")

    return obj_tql, tql_name, is_list, protocol


def sro_type_to_tql(sro_type: str,
                    sro_sub_type: str,
                    import_type:ImportType=default_import_type,
                    attack_object=False,
                    uses_relation=False,
                    is_procedure=False) -> [dict, str, list, str]:
    """ convert Stix object into a data model for processing

        Args:
            sro_type: the Stix2 type
            sro_sub_type: the relationship type
            import_type (): the type of import to use


        Returns:
            obj_tql : the dict of the tql properties
            tql_name : the typeql name of the object
            is_list: a list of all the porperties that are lists

    """
    # - list of property names that have values, and do not include False values
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    obj_tql = {}
    protocol = ""
    sro_tql_name = sro_type
    if sro_sub_type != "":
        sro_tql_name = sro_sub_type
    is_list = copy.deepcopy(auth["is_lists"]["sro"]["sro"])
    if sro_type == "sighting":
        is_list.extend(auth["is_lists"]["sro"]["sighting"])
    # If import_type is deaful None, then assign to default)
    if not import_type:
        import_type = default_import_type

    if auth['STIX21'] and not auth["ATT&CK"]:
        if stix_model.contains_data(sro_type):
            # dispatch specific stix properties plus later on, generic sdo properties
            protocol = "stix21"
            obj_tql = copy.deepcopy(stix_model.get_data(sro_type))
        elif auth["os-intel"] or auth["os-hunt"] and os_threat_model.contains_data(sro_type):
            # dispatch specific stix properties plus later on, generic sdo properties
            protocol = "os-threat"
            obj_tql = copy.deepcopy(os_threat_model.get_data(sro_type))
        else:
            logger.error(f'stixobj_type type {sro_type} not supported stix relation')
            return {}, "", []
    # - mitre attack_setting import
    elif auth['STIX21'] and auth["ATT&CK"]:
        if attack_object:
            is_list.extend(auth["is_lists"]["sro"]["attack"])
            protocol = "attack"
            attack_type = ''
            obj_tql = copy.deepcopy(attack_model.get_base("attack_base"))
            obj_tql.update(stix_model.get_data("relationship"))
            obj_tql.update({"x_mitre_platforms": "x-mitre-platforms"})
            # Convert from stix-type to attack-tql-entity
            for model in auth["reln"]["relations_sro_roles"]:
                if model["stix"] == sro_sub_type:
                    attack_type = model["typeql"]
                    #obj_tql.update(attack_models["data"][attack_type])
                    if uses_relation and is_procedure:
                        attack_type = "procedure"
                    break
            # Else log an error
            if not attack_type:
                logger.error(f'obj_type type {sro_type} not in attack type conversion dict, type_to_tql_name')
                return {}, "", []
            else:
                sro_tql_name = attack_type

        else:
            # its a Stix object, not an AT&CK one
            if stix_model.contains_data(sro_type):
                # dispatch specific stix properties plus mitre properties plus generic sdo properties
                protocol = "stix21"
                obj_tql = copy.deepcopy(stix_model.get_data(sro_type))
            elif os_threat_model.contains_data(sro_type) and auth["os-intel"] or auth["os-hunt"]:
                # dispatch specific stix properties plus later on, generic sdo properties
                protocol = "os-threat"
                obj_tql = copy.deepcopy(os_threat_model.get_data(sro_type))
            else:
                logger.error(f'obj_type type {sro_type} not in not any supported stix relation ')
                return {}, "", []

    else:
        logger.error(f'import type {import_type} not supported')
        return {}, "", []

    # - add on the generic sro properties
    obj_tql.update(stix_model.get_base("base_sro"))

    return obj_tql, sro_tql_name, is_list, protocol


def sco__type_to_tql(sco_type: str, import_type=default_import_type) -> [Dict[str, str], str, List[str], str]:
    """ convert Stix object into a data model for processing

        Args:
            sco_type (): the Stix2 sco type
            import_type (): the type of import to use

        Returns:
            obj_tql : a list of all properties
            obj_tql : the dict of the twl proeprties
            is_list: a list of all the porperties that are lists

    """
    # Based on import type setup observables
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    protocol = ""
    is_list = copy.deepcopy(auth["is_lists"]["sco"]["sco"])
    is_list.extend(auth["is_lists"]["sco"][sco_type])

    # - get the object-specific typeql names, sighting or relationship
    sco_tql_name = sco_type
    protocol = "stix21"
    obj_tql = copy.deepcopy(auth["objects"][sco_type])
    # - add on the generic sro properties
    obj_tql.update(stix_model.get_base("base_sco"))

    return obj_tql, sco_tql_name, is_list, protocol


def meta_type_to_tql(meta_type: str, import_type=default_import_type, attack_object=False) -> [Dict[str, str], str, List[str]]:
    """ convert Stix object into a data model for processing

        Args:
            meta_type (): the Stix2 meta type
            import_type (): the type of import to use

        Returns:
            obj_tql : a list of all properties
            obj_tql : the dict of the twl proeprties
            is_list: a list of all the porperties that are lists

    """
    # Based on import type setup observables
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    is_list = copy.deepcopy(auth["is_lists"]["sdo"]["sdo"])
    is_list.extend(auth["is_lists"]["meta"]["marking-definition"])
    meta_tql_name = meta_type
    obj_tql = copy.deepcopy(auth["objects"]["marking-definition"])
    # - add on the generic sro properties
    obj_tql.update(stix_model.get_base("base_sdo"))

    # - get the object-specific typeql names, sighting or relationship
    if attack_object:
        is_list.extend(auth["is_lists"]["sdo"]["attack"])
        obj_tql.update(attack_model.get_base("attack_base"))
        protocol = "attack"
        meta_tql_name = "attack-marking"
    else:
        protocol = "stix21"
        meta_tql_name = "statement-marking"

    return obj_tql, meta_tql_name, is_list, protocol



##################################################
# 2. Dispatch Dicts to convert between:
#      - Stix2 Object Property Name --> TypeQL Name
#      - Stix 2 Object Type --> Dict Name
###################################################

#---------------------------------------------------
# 2.1) Stix Domain Object Dicts
#---------------------------------------------------
sdo_typeql_dict = {
  "type" :  "stix-type",
  "spec_version" : "spec-version",
  "id"  : "stix-id",
  "created_by_ref"  : "",
  "created"  : "created",
  "modified" : "modified",
  "revoked"  : "revoked",
  "labels"  : "labels",
  "confidence"  : "confidence",
  "lang"  : "langs",
  "external_references"  : "",
  "object_marking_refs"  : "",
  "granular_markings"  : "",
  "extensions"  : ""
}

attack_pattern_typeql_dict = {
  "name": "name",
  "description": "description",
  "aliases": "aliases",
  "kill_chain_phases": ""
}

campaign_typeql_dict = {
  "name": "name",
  "description": "description",
  "aliases": "aliases",
  "first_seen": "first-seen",
  "last_seen": "last-seen",
  "objective": "objective"
}

grouping_typeql_dict = {
  "name": "name",
  "description": "description",
  "context": "context",
  "object_refs": ""
}

course_of_action_typeql_dict = {
  "name": "name",
  "description": "description",	
  "action": "action"
}

identity_typeql_dict = {
  "name" :  "name",
  "description" :  "description",
  "roles" :  "stix-role",
  "identity_class" :  "identity-class",
  "sectors" :  "sector",
  "contact_information" :  "contact-information"
}

incident_typeql_dict = {
  "name": "name",
  "description": "description",
}

indicator_typeql_dict = {
  "name": "name",
  "description": "description",
  "pattern_version": "pattern-version",
  "indicator_types": "indicator-type",
  "pattern": "pattern",
  "pattern_type": "pattern-type",
  "valid_from": "valid-from",
  "valid_until": "valid-until",
  "kill_chain_phases": ""
}

infrastructure_typeql_dict = {
  "name": "name",
  "description": "description",
  "infrastructure_types": "infrastructure-types",
  "aliases": "aliases",
  "kill_chain_phases": "",
  "first_seen": "first-seen",
  "last_seen": "last-seen"
}

intrusion_set_typeql_dict = {
  "name": "name",
  "description": "description",
  "aliases": "aliases",
  "first_seen": "first-seen",
  "last_seen": "last-seen",
  "goals": "goals",
  "resource_level": "resource-level",
  "primary_motivation": "primary-motivation",
  "secondary_motivations": "secondary-motivations"
}

location_typeql_dict = {
  "name": "name",
  "description": "description",
  "latitude": "latitude",
  "longitude": "longitude",
  "precision": "precision",
  "region": "region",
  "country": "country",
  "administrative_area": "administrative-area",
  "city": "city",
  "street_address": "street-address",
  "postal_code": "postal-code"
}

malware_typeql_dict = {
  "name": "name",	
  "description": "description",
  "malware_types": "malware-types",
  "is_family": "is-family",
  "aliases": "aliases",
  "kill_chain_phases": "",
  "first_seen": "first-seen",
  "last_seen": "last-seen",
  "operating_system_refs": "",
  "architecture_execution_envs": "architecture-execution-envs",
  "implementation_languages": "implementation-languages",
  "capabilities": "capabilities",
  "sample_refs": ""
}

malware_analysis_typeql_dict = {
  "product": "product",
  "version": "version",
  "host_vm_ref": "",
  "operating_system_ref": "",
  "installed_software_refs": "",
  "configuration_version": "configuration-version",
  "modules": "modules",
  "analysis_engine_version": "analysis-engine-version",
  "analysis_definition_version": "analysis-definition-version",
  "submitted": "submitted",
  "analysis_started": "analysis-started",
  "analysis_ended": "analysis-ended",
  "result_name": "result-name",
  "result": "result",
  "analysis_sco_refs": "",
  "sample_ref": ""
}

note_type_dict = {
  "abstract": "abstract",
  "content": "content",
  "authors": "authors",
  "object_refs": ""
}

observed_data_typeql_dict = {
  "first_observed": "first-observed",
  "last_observed": "last-observed",
  "number_observed": "number-observed",
  "object_refs": ""
}

opinion_typeql_dict = {
  "explanation": "explanation",
  "authors": "authors",
  "opinion": "opinion-enum",
  "object_refs": ""
}

report_typeql_dict = {
  "name": "name",
  "description": "description",
  "report_types": "report-type",
  "published": "published",
  "object_refs": ""
}

threat_actor_typeql_dict = {        
  "name"  : "name",
  "description"  : "description",
  "threat_actor_types"  : "threat-actor-type",
  "aliases"  : "aliases",
  "first_seen"  : "first-seen",
  "last_seen"  : "last-seen",
  "roles"  : "stix-role",
  "goals"  : "goals",
  "sophistication"  : "sophistication",
  "resource_level"  : "resource-level",
  "primary_motivation"  : "primary-motivation",
  "secondary_motivations"  : "secondary-motivations",
  "personal_motivations"  : "personal-motivations"        
}

tool_typeql_dict = {
  "name": "name",
  "description": "description",
  "tool_types": "tool-type",
  "aliases": "aliases",
  "kill_chain_phases": "",
  "tool_version": "tool-version"
}

vulnerability_typeql_dict = {
  "name": "name",
  "description": "description"
}

kill_chain_phases_typeql_dict = {
  "kill_chain_name": "kill-chain-name",
  "phase_name": "phase-name"
  
}


#---------------------------------------------------
# 2.2) Stix Cyber Observable Object Dicts
#---------------------------------------------------

sco_base_typeql_dict = {
  "type" :  "stix-type",
  "spec_version" : "spec-version",
  "id"  : "stix-id",
  "created_by_ref"  : "",
  "revoked"  : "revoked",
  "labels"  : "labels",
  "confidence"  : "confidence",
  "lang"  : "langs",
  "external_references"  : "",
  "object_marking_refs"  : "",
  "granular_markings"  : "",
  "defanged"  : "defanged",
  "extensions"  : ""
}

artifact_typeql_dict = {
  "mime_type": "mime-type",
  "payload_bin": "payload-bin",
  "url": "url-link",
  "hashes": "",
  "encryption_algorithm": "encryption-algorithm",
  "decryption_key": "decryption-key"
}

autonomous_system_typeql_dict = {
    "number": "number",
    "name": "name",
    "rir": "rir"
}

directory_typeql_dict = {
    "path": "path",
    "path_enc": "path-enc",
    "ctime": "ctime",
    "mtime": "mtime",
    "atime": "atime",
    "contains_refs": "",
}


domain_name_typeql_dict = {
    "value": "stix-value",
    "resolves_to_refs": ""
}

email_addr_typeql_dict = {
    "value": "stix-value",
    "display_name": "display-name",
    "belongs_to_ref": ""
}

email_message_typeql_dict = {
    "is_multipart": "is-multipart",
    "date": "date",
    "content_type": "content-type",
    "from_ref": "",
    "sender_ref": "",
    "to_refs": "",
    "cc_refs": "",
    "bcc_refs": "",
    "message_id": "message-id",
    "subject": "subject",
    "received_lines": "received-lines",
    "additional_header_fields": "",
    "body": "body",
    "body_multipart": "",
    "raw_email_ref": ""
}

file_typeql_dict = {
    "hashes": "",
    "size": "size",
    "name": "name",
    "name_enc": "name-enc",
    "magic_number_hex": "magic-number-hex",
    "mime_type": "mime-type",
    "ctime": "ctime",
    "dtime": "dtime",
    "atime": "atime",
    "parent_directory_ref": "",
    "contains_refs": "",
    "content_ref": ""
}

ipv4_addr_typeql_dict = {
    "value": "stix-value",
    "resolves_to_refs": "",
    "belongs_to_refs": ""
}

ipv6_addr_typeql_dict = {
    "value": "stix-value",
    "resolves_to_refs": "",
    "belongs_to_refs": ""
}

mac_addr_typeql_dict = {
    "value": "stix-value"
}

mutex_typeql_dict = {
    "name": "name"
}

network_traffic_typeql_dict = {
    "start": "start",
    "end": "end",
    "is_active": "is-active",
    "src_ref": "",
    "dst_ref": "",
    "src_port": "src-port",
    "dst_port": "dst-port",
    "protocols": "protocols",
    "src_byte_count": "src-byte-count",
    "dst_byte_count": "dst-byte-count",
    "src_packets": "src-packets",
    "dst_packets": "dst-packets",
    "ipfix": "",
    "src_payload_ref": "",
    "dst_payload_ref": "",
    "encapsulates_refs": "",
    "encapsulated_by_ref": ""    
}

process_typeql_dict = {
    "is_hidden": "is-hidden",
    "pid": "pid",
    "created_time": "created-time",
    "cwd": "cwd",
    "command_line": "command-line",
    "environment_variables": "",
    "opened_connection_refs": "",
    "creator_user_ref": "",
    "image_ref": "",
    "parent_ref": "",
    "child_refs": ""   
    
}

software_typeql_dict = {
    "name": "name",
    "cpe": "cpe",
    "swid": "swid",
    "languages": "language",
    "vendor": "vendor",
    "version": "version"
}


url_typeql_dict = {
    "value": "stix-value"
}

user_account_typeql_dict = {
    "user_id": "user-id",
    "credential": "credential",
    "account_login": "account-login",
    "account_type": "account-type",
    "display_name": "display-name",
    "is_service_account": "is-service-account",
    "is_privileged": "is-privileged",
    "can_escalate_privs": "can-escalate-privs",
    "is_disabled": "is-disabled",
    "account_created": "account-created",
    "account_expires": "account-expires",
    "credential_last_changed": "credential-last-changed",
    "account_first_login": "account-first-login",
    "account_last_login": "account-last-login"
}

windows_registry_key_typeql_dict = {
    "key": "attribute-key",
    "values": "",
    "modified_time": "modified-time",
    "creator_user_ref": "",
    "number_of_subkeys": "number-subkeys"
}

windows_registry_value_typeql_dict = {
    "name": "name",
    "data": "data",
    "data_type": "data-type"
}

x509_certificate_typeql_dict = {
    "is_self_signed": "is-self-signed",
    "hashes": "",
    "version": "version",
    "serial_number": "serial-number",
    "signature_algorithm": "signature-algorithm",
    "issuer": "issuer",
    "validity_not_before": "validity-not-before",
    "validity_not_after": "validity-not-after",
    "subject": "subject",
    "subject_public_key_algorithm": "subject-public-key-algorithm",
    "subject_public_key_modulus": "subject-public-key-modulus",
    "subject_public_key_exponent": "subject-public-key-exponent",
    "x509_v3_extensions": ""
}




#---------------------------------------------------
# 2.2b) SCO Extensions and Similar
#---------------------------------------------------

email_mime_part_typeql_dict = {
    "body": "body",
    "content_type": "content-type",
    "content_disposition": "content-disposition",
    "body_raw_ref": ""
}

archive_ext_typeql_dict = {
    "comment": "comment",
    "contains_refs": ""
}

ntfs_ext_typeql_dict = {
    "sid": "sid",
    "alternate_data_streams": ""
}

alternate_data_stream_ext_typeql_dict = {
    "name": "name",
    "size": "size",
    "hashes": ""
}

pdf_ext_typeql_dict = {
    "version": "version",
    "is_optimized": "is-optimized",
    "document_info_dict": "",
    "pdfid0": "pdfid0",
    "pdfid1": "pdfid1"
}

raster_image_ext_typeql_dict = {
    "image_height": "image-height",
    "image_width": "image-width",
    "exif_tags": "",
    "bits_per_pixel": "bits-per-pixel"
}

windows_pebinary_ext_typeql_dict = {
    "pe_type": "pe-type",
    "imphash": "imphash",
    "machine_hex": "machine-hex",
    "number_of_sections": "number-of-sections",
    "time_date_stamp": "time-date-stamp",
    "pointer_to_symbol_table_hex": "pointer-to-symbol-table-hex",
    "number_of_symbols": "number-of-symbols",
    "size_of_optional_header": "size-of-optional-header",
    "characteristics_hex": "characteristics-hex",
    "file_header_hashes": "",
    "optional_header": "",
    "sections": ""
}

windows_optional_header_ext_typeql_dict = {
    "magic_hex": "magic-hex",
    "major_linker_version": "major-linker-version",
    "minor_linker_version": "minor-linker-version",
    "size_of_code": "size-of-code",
    "size_of_initialized_data": "size-of-initialized-data",
    "size_of_uninitialized_data": "size-of-uninitialized-data",
    "address_of_entry_point": "address-of-entry-point",
    "base_of_code": "base-of-code",
    "base_of_data": "base-of-data",
    "image_base": "image-base",
    "section_alignment": "section-alignment",
    "file_alignment": "file-alignment",
    "major_os_version": "major-os-version",
    "minor_os_version": "minor-os-version",
    "major_image_version": "major-image-version",
    "minor_image_version": "minor-image-version",
    "major_subsystem_version": "major-subsystem-version",
    "minor_subsystem_version": "minor-subsystem-version",
    "win32_version_value_hex": "win32-version-value-hex",
    "size_of_image": "size-of-image",
    "size_of_headers": "size-of-headers",
    "checksum_hex": "checksum-hex",
    "subsystem_hex": "subsystem-hex",
    "dll_characteristics_hex": "dll-characteristics-hex",
    "size_of_stack_reserve": "size-of-stack-reserve",
    "size_of_stack_commit": "size-of-stack-commit",
    "size_of_heap_reserve": "size-of-heap-reserve",
    "size_of_heap_commit": "size-of-heap-commit",
    "loader_flags_hex": "loader-flags-hex",
    "number_of_rva_and_sizes": "number-of-rva-and-sizes",
    "hashes": ""
}

windows_pe_section_ext_typeql_dict = {
    "name": "name",
    "size": "size",
    "entropy": "entropy",
    "hashes": ""
}

HTTP_request_ext_typeql_dict = {
    "request_method": "request-method",
    "request_value": "request-value",
    "request_version": "request-version",
    "request_header": "",
    "message_body_length": "message-body-length",
    "message_body_data_ref": ""
}

icmp_ext_typeql_dict = {
    "icmp_type_hex": "icmp-type-hex",
    "icmp_code_hex": "icmp-code-hex"
}

socket_ext_typeql_dict = {
    "address_family": "address-family",
    "is_blocking": "is-blocking",
    "is_listening": "is-listening",
    "options": "",
    "socket_type": "socket-type",
    "socket_description": "socket-description",
    "socket_handle": "socket-handle"
}


tcp_ext_typeql_dict = {
    "src_flags_hex": "src-flags-hex",
    "dst_flags_hex": "dst-flags-hex"
}

windows_process_ext_typeql_dict = {
    "aslr_enabled": "aslr-enabled",
    "dep_enabled": "dep-enabled",
    "priority": "priority",
    "owner_sid": "owner-sid",
    "window_title": "window-title",
    "startup_info": "",
    "integrity_level": "integrity-level"
}

windows_service_ext_typeql_dict = {
    "service_name": "service-name",
    "descriptions": "description",
    "display_name": "display-name",
    "group_name": "group-name",
    "start_type": "start-type",
    "service_dll_refs": "",
    "service_type": "service-type",
    "service_status": "service-status"
}

unix_account_ext_typeql_dict = {
    "gid": "gid",
    "groups": "unix-group",
    "home_dir": "home-dir",
    "shell": "shell"
}

x509_v3_ext_typeql_dict = {
    "basic_constraints": "basic-constraints",
    "name_constraints": "name-constraints",
    "policy_constraints": "policy-constraints",
    "key_usage": "key-usage",
    "extended_key_usage": "extended-key-usage",
    "subject_key_identifier": "subject-key-identifier",
    "authority_key_identifier": "authority-key-identifier",
    "subject_alternative_name": "subject-alternative-name",
    "issuer_alternative_name": "issuer-alternative-name",
    "subject_directory_attributes": "subject-directory-attributes",
    "crl_distribution_points": "crl-distribution-points",
    "inhibit_any_policy": "inhibit-any-policy",
    "private_key_usage_period_not_before": "private-key-usage-period-not-before",
    "private_key_usage_period_not_after": "private-key-usage-period-not-after",
    "certificate_policies": "certificate-policies",
    "policy mapping": "policy-mapping"
}



ext_typeql_dict_list = [
    {   "stix": "archive-ext",
        "dict": archive_ext_typeql_dict,
        "object": "archive-ext",
        "relation": "archive-extension", 
        "owner": "file",
        "pointed-to": "an-archive"},
    {   "stix": "ntfs-ext",
        "dict": ntfs_ext_typeql_dict,
        "object": "ntfs-ext",
        "relation": "ntfs-extension", 
        "owner": "file",
        "pointed-to": "ntfs"},
    {   "stix": "alternate_data_streams",
        "dict": alternate_data_stream_ext_typeql_dict,
        "object": "alternate-data-stream",
        "relation": "alt-data-streams",
        "owner": "ntfs-ext",
        "pointed-to": "alt-data-stream"},
    {   "stix": "pdf-ext",
        "dict": pdf_ext_typeql_dict,
        "object": "pdf-ext",
        "relation": "pdf-extension", 
        "owner": "file",
        "pointed-to": "pdf"},
    {   "stix": "raster-image-ext",
        "dict": raster_image_ext_typeql_dict,
        "object": "raster-image-ext",
        "relation": "raster-image-extension", 
        "owner": "file",
        "pointed-to": "image"},
    {   "stix": "windows-pebinary-ext",
        "dict": windows_pebinary_ext_typeql_dict,
        "object": "windows-pebinary-ext",
        "relation": "windows-pebinary-extension", 
        "owner": "file",
        "pointed-to": "pebinary"},
    {   "stix": "windows-pe-section-type",
        "dict": windows_pe_section_ext_typeql_dict,
        "object": "windows-pe-section",
        "relation": "sections", 
        "owner": "pebinary",
        "pointed-to": "pe-section"},
    {   "stix": "http-request-ext",
        "dict": HTTP_request_ext_typeql_dict,
        "object": "http-request-ext",
        "relation": "http-request-extension", 
        "owner": "traffic",
        "pointed-to": "request"},
    {   "stix": "icmp-ext",
        "dict": icmp_ext_typeql_dict,
        "object": "icmp-ext",
        "relation": "icmp-extension", 
        "owner": "traffic",
        "pointed-to": "icmp"},
    {   "stix": "socket-ext",
        "dict": socket_ext_typeql_dict,
        "object": "socket-ext",
        "relation": "socket-extension", 
        "owner": "traffic",
        "pointed-to": "socket"},
    {   "stix": "tcp-ext",
        "dict": tcp_ext_typeql_dict,
        "object": "tcp-ext",
        "relation": "tcp-extension", 
        "owner": "traffic",
        "pointed-to": "tcp"},
    {   "stix": "windows-process-ext",
        "dict": windows_process_ext_typeql_dict,
        "object": "windows-process-ext",
        "relation": "windows-process-extension", 
        "owner": "process",
        "pointed-to": "win-process"},
    {   "stix": "windows-service-ext",
        "dict": windows_service_ext_typeql_dict,
        "object": "windows-service-ext",
        "relation": "windows-service-extension", 
        "owner": "process",
        "pointed-to": "win-service"},
    {   "stix": "unix-account-ext",
        "dict": unix_account_ext_typeql_dict,
        "object": "unix-account-ext",
        "relation": "unix-account-extension", 
        "owner": "account",
        "pointed-to": "unix"},
    {   "stix": "x509_v3_extensions",
        "dict": x509_v3_ext_typeql_dict,
        "object": "x509-v3-extension",
        "relation": "v3-extensions", 
        "owner": "cert",
        "pointed-to": "v3-extension"},
    {   "stix": "optional_header",
        "dict": windows_optional_header_ext_typeql_dict,
        "object": "windows-pe-optional-header-type",
        "relation": "optional-headers", 
        "owner": "pebinary",
        "pointed-to": "optional-header"}
]
  
  
  

#---------------------------------------------------
# 2.3) Marking Definition
#---------------------------------------------------

marking_typeql_dict = {
  "type": "stix-type",
  "id": "stix-id",
  "spec_version": "spec-version",
  "created": "created",
  "name": "name",
  "statement": "statement"
  
}

ext_ref_typeql_dict = {
  "source_name": "source-name",
  "description": "description",
  "url": "url-link",
  "hashes": "",
  "external_id": "external-id"	
}

hash_typeql_dict = {
  "MD5": "md-5",
  "SHA-1": "sha-1",
  "SHA-256": "sha-256",
  "SHA-512": "sha-512",
  "SHA3-256": "sha3-256",
  "SHA3-512": "sha3-512",
  "SSDEEP": "ssdeep",
  "TLSH": "tlsh"
}



#---------------------------------------------------
# 2.4) Stix type_ql_relationhip Object Dict and TypeQL Roles List of Dicts
#---------------------------------------------------

sro_base_typeql_dict = {
  "type" :  "stix-type",
  "spec_version" :  "spec-version",
  "id" :  "stix-id",
  "created_by_ref" :  "",
  "created" :  "created",
  "modified" :  "modified",
  "revoked" :  "revoked",
  "labels" :  "labels",
  "confidence" :  "confidence",
  "lang" :  "langs",
  "external_references" :  "",
  "object_marking_refs" :  "",
  "granular_markings" :  "",
  "extensions" :  ""
}

relationship_typeql_dict = {        
  "relationship_type" :  "relationship-type",
  "description" :  "description",
  "source_ref" :  "",
  "target_ref" :  "",
  "start_time" :  "start-time",
  "stop_time" :  "stop-time"        
}

sighting_typeql_dict = {
  "description" :  "description",
  "first_seen" :  "first-seen",
  "last_seen" :  "last-seen",
  "count" :  "count",
  "sighting_of_ref" :  "",
  "observed_data_refs" :  "",
  "where_sighted_refs" :  "",
  "summary" :  "summary"
}
    

stix_rel_roles = [
 {   "stix": "delivers",    "typeql": "delivers",    "source": "delivering",    "target": "delivered" }, 
 {   "stix": "targets",    "typeql": "targets",   "source": "targetter",   "target": "targetted" }, 
 {   "stix": "uses",   "typeql": "uses",   "source": "used-by",   "target": "used" }, 
 {   "stix": "attributed-to",   "typeql": "attributed-to",   "source": "result",   "target": "fault-of" }, 
 {   "stix": "compromises",   "typeql": "compromises",   "source": "compromising",   "target": "compromised" }, 
 {   "stix": "originates-from", "typeql": "originates-from",   "source": "originating",   "target": "originated-from" }, 
 {   "stix": "investigates",   "typeql": "investigates",   "source": "investigating",   "target": "investigated" }, 
 {   "stix": "mitigates",   "typeql": "mitigates",   "source": "mitigator",   "target": "mitigated" }, 
 {   "stix": "located-at",   "typeql": "located-at",   "source": "locating",   "target": "located" }, 
 {   "stix": "indicates",   "typeql": "indicates",   "source": "indicating",   "target": "indicated" }, 
 {   "stix": "based-on",   "typeql": "based-on",   "source": "basing-on",   "target": "basis" }, 
 {   "stix": "communicates-with",   "typeql": "communicates-with",   "source": "communicating",   "target": "communicated" }, 
 {   "stix": "consists-of",   "typeql": "consist",   "source": "consisting",   "target": "consisted" }, 
 {   "stix": "controls",   "typeql": "control",   "source": "controlling",   "target": "controlled" }, 
 {   "stix": "has",   "typeql": "have",   "source": "having",   "target": "had" }, 
 {   "stix": "hosts",   "typeql": "hosts",   "source": "hosting",   "target": "hosted" }, 
 {   "stix": "owns",   "typeql": "ownership",   "source": "owning",   "target": "owned" }, 
 {   "stix": "authored-by",   "typeql": "authored-by",   "source": "authoring",   "target": "authored" }, 
 {   "stix": "beacons-to",   "typeql": "beacon",   "source": "beaconing-to",   "target": "beaconed-to" }, 
 {   "stix": "exfiltrate-to",   "typeql": "exfiltrate",   "source": "exfiltrating-to",   "target": "exfiltrated-to" }, 
 {   "stix": "downloads",   "typeql": "download",   "source": "downloading",   "target": "downloaded" }, 
 {   "stix": "drops",   "typeql": "drop",   "source": "dropping",   "target": "dropped" }, 
 {   "stix": "exploits",   "typeql": "exploit",   "source": "exploiting",   "target": "exploited" }, 
 {   "stix": "variant-of",   "typeql": "variant",   "source": "variant-source",   "target": "variant-target" }, 
 {   "stix": "characterizes",   "typeql": "characterise",   "source": "characterising",   "target": "characterised" }, 
 {   "stix": "analysis-of",   "typeql": "av-analysis",   "source": "analysing",   "target": "analysed" }, 
 {   "stix": "static-analysis-of",   "typeql": "static-analysis",   "source": "analysing",   "target": "analysed" }, 
 {   "stix": "dynamic-analysis-of",   "typeql": "dynamic-analysis",   "source": "analysing",   "target": "analysed" }, 
 {   "stix": "impersonates",   "typeql": "impersonate",   "source": "impersonating",   "target": "impersonated" }
]

embedded_relations_typeql = [
  {"rel": "object_refs", "owner": "object", "pointed-to": "referred", "typeql": "obj-ref"},
  {"rel": "created_by_ref", "owner": "created", "pointed-to": "creator", "typeql": "created-by"},
  {"rel": "object_marking_refs", "owner": "marked", "pointed-to": "marking", "typeql": "object-marking"},
  {"rel": "sample_refs", "owner": "sample-for", "pointed-to": "sco-sample", "typeql": "malware-sample"},
  {"rel": "sample_ref", "owner": "sample-for", "pointed-to": "sco-sample", "typeql": "malware-analysis-sample"},
  {"rel": "host_vm_ref", "owner": "object", "pointed-to": "env", "typeql": "host-vm-ref"},
  {"rel": "operating_system_ref", "owner": "object", "pointed-to": "env", "typeql": "operating-system"},
  {"rel": "installed_software_refs", "owner": "object", "pointed-to": "env", "typeql": "installed-software"},
  {"rel": "contains_refs", "owner": "container", "pointed-to": "contained", "typeql": "directory-contains"},
  {"rel": "parent_directory_ref", "owner": "contained", "pointed-to": "container", "typeql": "directory-parent"},
  {"rel": "resolves_to_refs", "owner": "resolve", "pointed-to": "resolves-to", "typeql": "resolves"},
  {"rel": "belongs_to_ref", "owner": "belonged", "pointed-to": "belongs-to", "typeql": "belongs"},
  {"rel": "belongs_to_refs", "owner": "belonged", "pointed-to": "belongs-to", "typeql": "belongs-to-autonomous"},
  {"rel": "analysis_sco_refs", "owner": "object", "pointed-to": "env", "typeql": "captured-objects"},
  {"rel": "raw_email_ref", "owner": "email", "pointed-to": "binary", "typeql": "raw-email-references"},
  {"rel": "from_ref", "owner": "email", "pointed-to": "email-address", "typeql": "from-email"},
  {"rel": "sender_ref", "owner": "email", "pointed-to": "email-address", "typeql": "sender-email"},
  {"rel": "to_refs", "owner": "email", "pointed-to": "email-address", "typeql": "to-email"},
  {"rel": "cc_refs", "owner": "email", "pointed-to": "email-address", "typeql": "cc-email"},
  {"rel": "bcc_refs", "owner": "email", "pointed-to": "email-address", "typeql": "bcc-email"},
  {"rel": "body_raw_ref", "owner": "containing-mime", "pointed-to": "non-textual", "typeql": "body-raw-references"},
  {"rel": "content_ref", "owner": "containing-file", "pointed-to": "content", "typeql": "content-file"},
  {"rel": "src_ref", "owner": "traffic", "pointed-to": "source", "typeql": "traffic-src"},
  {"rel": "src_payload_ref", "owner": "traffic", "pointed-to": "source", "typeql": "payload-src"},
  {"rel": "dst_ref", "owner": "traffic", "pointed-to": "destination", "typeql": "traffic-dst"},
  {"rel": "dst_payload_ref", "owner": "traffic", "pointed-to": "payload", "typeql": "payload-dst"},
  {"rel": "encapsulates_refs", "owner": "container", "pointed-to": "contained", "typeql": "encapsulate"},
  {"rel": "encapsulated_by_ref", "owner": "contained", "pointed-to": "container", "typeql": "encapsulated"},
  {"rel": "message_body_data_ref", "owner": "HTPP-message", "pointed-to": "container", "typeql": "HTTP-body-data"},
  {"rel": "opened_connection_refs", "owner": "process", "pointed-to": "opened-connection", "typeql": "open-connections"},
  {"rel": "creator_user_ref", "owner": "created", "pointed-to": "creator", "typeql": "user-created-by"},
  {"rel": "image_ref", "owner": "process", "pointed-to": "executed-image", "typeql": "process-image"},
  {"rel": "parent_ref", "owner": "process", "pointed-to": "parent", "typeql": "process-parent"},
  {"rel": "child_refs", "owner": "process", "pointed-to": "child", "typeql": "process-child"},
  {"rel": "service_dll_refs", "owner": "process", "pointed-to": "loaded-dll", "typeql": "service-dll"}
]



key_value_typeql_list = [
    {
        "name": "additional_header_fields", 
        "typeql": "additional-header", 
        "owner": "email", 
        "pointed_to": "item", 
        "key": "header-key",
        "value": "header-value"
    },{
        "name": "document_info_dict", 
        "typeql": "doc-info", 
        "owner": "pdf", 
        "pointed_to": "info", 
        "key": "doc-key",
        "value": "doc-value"
    },{
        "name": "exif_tags", 
        "typeql": "EXIF-tags", 
        "owner": "image", 
        "pointed_to": "info", 
        "key": "EXIF-key",
        "value": "EXIF-value"
    },{
        "name": "ipfix", 
        "typeql": "IPFIX-store", 
        "owner": "traffic", 
        "pointed_to": "item", 
        "key": "IPFIX-key",
        "value": "IPFIX-value"
    },{
        "name": "request_header", 
        "typeql": "HTTP-header", 
        "owner": "request", 
        "pointed_to": "header", 
        "key": "HTTP-key",
        "value": "HTTP-value"
    },{
        "name": "options", 
        "typeql": "socket-options", 
        "owner": "socket", 
        "pointed_to": "option", 
        "key": "socket-key",
        "value": "socket-value"
    },{
        "name": "environment_variables", 
        "typeql": "environment-variables", 
        "owner": "process", 
        "pointed_to": "env-variable", 
        "key": "environment-key",
        "value": "environment-value"
    },{
        "name": "startup_info", 
        "typeql": "startup-info", 
        "owner": "process", 
        "pointed_to": "option", 
        "key": "startup-key",
        "value": "startup-value"
    }
]



list_of_object_typeql = [
    {
        "name": "body_multipart", 
        "typeql": "body-multipart", 
        "typeql_props": email_mime_part_typeql_dict, 
        "owner": "email", 
        "pointed_to": "mime-part", 
        "object": "email-mime-part"
    },{
        "name": "external_references", 
        "typeql": "external-references", 
        "typeql_props": ext_ref_typeql_dict, 
        "owner": "referencing", 
        "pointed_to": "referenced", 
        "object": "external-reference"
    },{
        "name": "kill_chain_phases", 
        "typeql": "kill-chain-usage", 
        "typeql_props": kill_chain_phases_typeql_dict, 
        "owner": "kill-chain-used", 
        "pointed_to": "kill-chain-using", 
        "object": "kill-chain-phase"
    },{
        "name": "alternate_data_streams", 
        "typeql": "alt-data-streams",
        "typeql_props": alternate_data_stream_ext_typeql_dict, 
        "owner": "ntfs-ext", 
        "pointed_to": "alt-data-stream",
        "object": "alternate-data-stream"
    },{
        "name": "sections",
        "typeql_props": windows_pe_section_ext_typeql_dict,
        "object": "windows-pe-section",
        "typeql": "sections", 
        "owner": "pebinary",
        "pointed_to": "pe-section"
    },{
        "name": "values",
        "typeql_props": windows_registry_value_typeql_dict,
        "object": "windows-registry-value-type",
        "typeql": "reg-val", 
        "owner": "reg-key",
        "pointed_to": "reg-value"
    }
]
 

    

#---------------------------------------------------
# 2.5) Object to Dict Mapping
#---------------------------------------------------

dispatch_stix = {
    
    "attack-pattern" :  attack_pattern_typeql_dict,
    "campaign" :  campaign_typeql_dict,
    "course-of-action" :  course_of_action_typeql_dict,
    "grouping" :  grouping_typeql_dict,
    "identity": identity_typeql_dict,
    "incident": incident_typeql_dict,
    "indicator": indicator_typeql_dict,
    "infrastructure": infrastructure_typeql_dict,
    "intrusion-set": intrusion_set_typeql_dict,
    "location": location_typeql_dict,
    "malware": malware_typeql_dict,
    "malware-analysis": malware_analysis_typeql_dict,
    "note" :  note_type_dict,
    "observed-data" :  observed_data_typeql_dict,
    "opinion" :  opinion_typeql_dict,
    "report" :  report_typeql_dict,
    "threat-actor": threat_actor_typeql_dict,
    "tool": tool_typeql_dict,
    "vulnerability": vulnerability_typeql_dict,
    "relationship": relationship_typeql_dict,
    "sighting": sighting_typeql_dict,
    "artifact": artifact_typeql_dict,
    "autonomous-system": autonomous_system_typeql_dict,
    "directory": directory_typeql_dict,
    "domain-name": domain_name_typeql_dict,
    "email-addr": email_addr_typeql_dict,
    "email-message": email_message_typeql_dict,
    "file": file_typeql_dict,
    "ipv4-addr": ipv4_addr_typeql_dict,
    "ipv6-addr": ipv6_addr_typeql_dict,
    "mac-addr": mac_addr_typeql_dict,
    "mutex": mutex_typeql_dict,
    "network-traffic": network_traffic_typeql_dict,
    "process": process_typeql_dict,
    "software": software_typeql_dict,
    "url": url_typeql_dict,
    "user-account": user_account_typeql_dict,
    "windows-registry-key": windows_registry_key_typeql_dict,
    "windows-registry-value-type": windows_registry_value_typeql_dict,
    "x509-certificate": x509_certificate_typeql_dict,
    "external-reference": ext_ref_typeql_dict,
    "email-mime-part": email_mime_part_typeql_dict,
    "archive-ext":  archive_ext_typeql_dict,
    "ntfs-ext":  ntfs_ext_typeql_dict,
    "alternate-data-stream": alternate_data_stream_ext_typeql_dict,
    "pdf-ext": pdf_ext_typeql_dict,
    "raster-image-ext": raster_image_ext_typeql_dict,
    "windows-pebinary-ext": windows_pebinary_ext_typeql_dict,
    "windows-pe-optional-header-type": windows_optional_header_ext_typeql_dict,
    "windows-pe-section": windows_pe_section_ext_typeql_dict,
    "http-request-ext": HTTP_request_ext_typeql_dict,
    "icmp-ext": icmp_ext_typeql_dict,
    "socket-ext": socket_ext_typeql_dict,
    "tcp-ext": tcp_ext_typeql_dict,
    "unix-account-ext": unix_account_ext_typeql_dict,
    "windows-process-ext": windows_process_ext_typeql_dict,
    "windows-service-ext": windows_service_ext_typeql_dict,
    "kill-chain-phase": kill_chain_phases_typeql_dict,
    "x509-v3-extension": x509_v3_ext_typeql_dict
}

sdo_obj = [
    "attack-pattern",
    "campaign" ,
    "course-of-action",
    "grouping",
    "identity",
    "incident",
    "indicator",
    "infrastructure",
    "intrusion-set",
    "location",
    "malware",
    "malware-analysis",
    "note",
    "observed-data" ,
    "opinion",
    "report",
    "threat-actor",
    "tool",
    "vulnerability"
]

sro_obj = [    
    "relationship",
    "sighting",
    "delivers",
    "targets",
    "uses",
    "attributed-to",
    "compromises",
    "originates-from",
    "investigates",
    "mitigates",
    "located-at",
    "indicates",
    "based-on",
    "communicates-with",
    "consist",
    "control",
    "have",
    "hosts",
    "ownership",
    "authored-by",
    "beacon",
    "exfiltrate",
    "download",
    "drop",
    "exploit",
    "variant",
    "characterise",
    "impersonate",
    "av-analysis",
    "static-analysis",
    "dynamic-analysis",
    "remediation"
]

sco_obj =[   
    "artifact",
    "autonomous-system",
    "directory",
    "domain-name",
    "email-addr",
    "email-message",
    "file",
    "ipv4-addr",
    "ipv6-addr",
    "mac-addr",
    "mutex",
    "network-traffic",
    "process",
    "software",
    "url",
    "user-account",
    "windows-registry-key",
    "windows-registry-value-type",
    "x509-certificate"
]    
    
meta_obj = [
    "marking-definition",
    "tlp-white",
    "tlp-green",
    "tlp-amber",
    "tlp-red",
    "statement-marking"
]
    

dispatch_attack = {}

extensions_only = [
    "archive-extension",
    "ntfs-extension",
    "pdf-extension",
    "raster-image-extension",
    "windows-pebinary-extension",
    "http-request-extension",
    "icmp-extension",
    "socket-extension",
    "tcp-extension",
    "unix-account-extension",
    "windows-process-extension",
    "windows-service-extension"
]

object_is_list = {
    "external-reference": [],
    "email-mime-part": [],
    "archive-ext":  ["contains_refs"],
    "ntfs-ext":  ["alternate_data_streams"],
    "alternate-data-stream": [],
    "pdf-ext": [],
    "raster-image-ext": [],
    "windows-pebinary-ext": ["sections"],
    "windows-pe-optional-header-type": [],
    "windows-pe-section": [],
    "http-request-ext": [],
    "icmp-ext": [],
    "socket-ext": [],
    "tcp-ext": [],
    "unix-account-ext": ["groups"],
    "windows-process-ext": [],
    "windows-service-ext": ["descriptions", "service_dll_refs"],
    "kill-chain-phase": [],
    "windows-registry-value-type": [],
    "x509-v3-extension": []
}

sdo_is_list = {
    "sdo": ["labels","external_references", "object_marking_refs", "granular_markings"],
    "attack-pattern": ["aliases", "kill_chain_phases"],
    "campaign": ["aliases"],
    "course-of-action": [],
    "grouping": ["object_refs"],
    "identity": ["roles", "sectors"],
    "incident": [],
    "indicator": ["indicator_types", "kill_chain_phases"],
    "infrastructure": ["infrastructure_types", "aliases", "kill_chain_phases"],
    "intrusion-set": ["aliases", "goals", "secondary_motivations" ],
    "location": [],
    "malware": ["malware_types", "kill_chain_phases","aliases", "operating_system_refs", "architecture_execution_envs", "implementation_languages", "capabilities", "sample_refs"],
    "malware-analysis": ["installed_software_refs", "modules", "analysis_sco_refs"],
    "note": ["authors", "object_refs"],
    "observed-data": ["object_refs"],
    "opinion": ["authors", "object_refs"],
    "report": ["report_types", "object_refs"],
    "threat-actor": ["threat_actor_types", "aliases", "roles", "goals","resource-level", "secondary_motivations", "personal_motivations"],
    "tool": ["tool_types", "kill_chain_phases", "aliases"],
    "vulnerability": []
}

sro_is_list = {
    "sro": ["labels","external_references", "object_marking_refs", "granular_markings"],
    "sighting": [ "observed_data_refs", "where_sighted_refs"]
}

sco_is_list = {
    "sco": ["labels","external_references", "object_marking_refs", "granular_markings"],
    "artifact": [],
    "autonomous-system": [],
    "directory": ["contains_refs"],
    "domain-name": ["resolves_to_refs"],
    "email-addr": [],
    "email-message": ["to_refs", "cc_refs", "bcc_refs", "received_lines", "body_multipart"],
    "email-mime-part": [],
    "file": ["contains_refs"],    
    "archive-ext":  ["contains_refs"],
    "ntfs-ext":  ["alternate_data_streams"],
    "alternate-data-stream-type": [],
    "pdf-ext": [],
    "raster-image-ext": [],
    "windows-pebinary-ext": ["sections"],
    "windows-pe-optional-header-type": [],
    "windows-pe-section-type": [],
    "ipv4-addr": ["resolves_to_refs", "belongs_to_refs"],
    "ipv6-addr": ["resolves_to_refs", "belongs_to_refs"],
    "mac-addr": [],
    "mutex": [],
    "network-traffic": [ "protocols", "encapsulates_refs"],
    "http-request-ext": [],
    "icmp-ext": [],
    "socket-ext": [],
    "tcp-ext": [],
    "process": ["opened_connection_refs", "child_refs"],
    "windows-process-ext": [],
    "windows-service-ext": ["descriptions", "service_dll_refs"],
    "software": ["languages"],
    "url": [],
    "user-account": [],
    "unix-account-ext": ["groups"],
    "windows-registry-key": ["values"],
    "windows-registry-value-type": [],
    "x509-certificate": [],
    "x509-v3-extensions-type": []
}


import functools
import hashlib

from stixorm.module.authorise import authorised_mappings
from stixorm.module.typedb_lib.factories.import_type_factory import ImportType


class AuthFactory:

    def __init__(self):
        self.auths_by_import_type = {}


    def get_auth_for_import(self,
                            import_type: ImportType):
        json_string = import_type.json()
        hashed_string = hashlib.sha256(json_string.encode()).hexdigest()
        if hashed_string in self.auths_by_import_type.keys():
            return self.auths_by_import_type[hashed_string]
        else:
            authorised_mapping = authorised_mappings(import_type)
            self.auths_by_import_type[hashed_string] = authorised_mapping
            return authorised_mapping

@functools.lru_cache(maxsize=None)
def get_auth_factory_instance() -> AuthFactory:
    return AuthFactory()


import json
import pathlib
import traceback
from typing import List
import copy

from stixorm.module.authorise import authorised_mappings
from stixorm.module.parsing.conversion_decisions import sdo_type_to_tql, sro_type_to_tql, sco__type_to_tql, meta_type_to_tql
from stixorm.module.orm.export_utilities import convert_ans_to_res
import logging

from stixorm.module.typedb_lib.factories.auth_factory import get_auth_factory_instance
from stixorm.module.typedb_lib.factories.import_type_factory import ImportType

logger = logging.getLogger(__name__)


###################################################################################################
#
#    TypeQL to Stix Mapping
#
###################################################################################################


# --------------------------------------------------------------------------------------------------------
#  Overview:
#     1. Convert TypeQL Ans to Res, using the transaction
#     2. Convert Res to Stix, creating first the dict, then parsing the dict to Stix object
# --------------------------------------------------------------------------------------------------------


def convert_ans_to_stix(query, answer_iterator, r_tx, import_type: ImportType):
    """
        High level function to convert the typedb_lib return into a Stix object.
        Firstly, drive the grpc to make an intermediate format, then convert that to a Stix dict
    Args:
        answer_iterator (): the returned iterator from the typedb_lib query
        r_tx (): the transaction
        import_type (): the type of import STIX21 or ATT&CK

    Returns:
        stix_dict {}: a dict containing the stix object
    """
    res = convert_ans_to_res(answer_iterator, r_tx, import_type)
    path = pathlib.Path(__file__).parent.joinpath("export_test.json")
    #with open(str(path), 'w') as outfile:
    #    json.dump(res, outfile)
    logger.debug(f'got res, now for stix')
    stix_dict = convert_res_to_stix(res, import_type)
    logger.debug((f'got stix now for object'))
    logger.debug("=========================================")
    json_object = json.dumps(stix_dict, indent=4)
    logger.debug(json_object)
    logger.debug("=========================================")
    return json_object


# --------------------------------------------------------------------------------------------------------
#  2. Convert Res to Stix
# --------------------------------------------------------------------------------------------------------

def convert_res_to_stix(res: List[dict], import_type: ImportType):
    """
        High level function to conver the intermediate form into a stix dict
    Args:
        res (): the intermediate form
        import_type (): the type of import "STIX21" or "ATT&CK"

    Returns:
        stix_dict {}: a dict containing the stix object
    """

    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    stix_dict = {}
    for obj in res:
        obj_type = obj["T_name"]
        tql_type = obj["type"]
        if obj_type in auth["tql_types"]["sdo"]:
            stix_dict = make_sdo(obj, import_type)
        elif obj_type in auth["tql_types"]["sco"]:
            stix_dict = make_sco(obj, import_type)
        elif tql_type in auth["tql_types"]["sro"] or tql_type == 'relationship':
            stix_dict = make_sro(obj, import_type)
        elif obj_type in auth["tql_types"]["meta"] or obj_type == "statement-marking":
            stix_dict = make_meta(obj, import_type)
        else:
            logger.error(f'Unknown object type: {obj}')

    logger.debug(f'\n\n')
    logger.debug(f'stix dict -> {stix_dict}\n')
    return stix_dict


def make_sdo(res, import_type: ImportType):
    """
        High-level function to convert intermediate format into a Stix domain object
    Args:
        res (): intermediate format
        import_type (): the type of import "STIX21" or "ATT&CK"

    Returns:
        stix_dict {}: a dict containing the stix object
    """
    try:
        auth_factory = get_auth_factory_instance()
        auth = auth_factory.get_auth_for_import(import_type)
        stix_dict = {}
        # 2.A) get the typeql properties and relations
        sdo_tql_name = res["T_name"]
        sdo_type = ""
        for model in auth["conv"]["sdo"]:
            if model["typeql"] == sdo_tql_name:
                sdo_type = model["type"]
        props = res["has"]
        relns = res["relns"]
        attack_object = False
        sub_technique = False
        for prop in props:
            if prop["typeql"] == "x-mitre-version":
                attack_object = True
            if prop["typeql"] == "x-mitre-is-subtechnique" and prop["value"] is True:
                sub_technique = True

        obj_tql, sdo_tql_name, is_list, protocol = sdo_type_to_tql(sdo_type, import_type, attack_object, sub_technique)

        #logger.debug(f"obj tql -> {obj_tql}\n sdo tql name -> {sdo_tql_name}")
        # 2.B) get the is_list list, the list of properties that are lists for that object
        #is_list = auth["is_lists"]["sdo"]["sdo"] + auth["is_lists"]["sdo"][sdo_tql_name]
        # 3.A) add the properties onto the the object
        stix_dict = make_properties(props, obj_tql, stix_dict, is_list)
        #logger.debug(f'sdo, add properties, stix_dict -> {stix_dict}\n')
        # 3.B) add the relations onto the object
        stix_dict = make_relations(relns, obj_tql, stix_dict, is_list, sdo_tql_name, import_type)
        #logger.debug(f'sdo, add relations, , stix_dict -> {stix_dict}\n')
        # 4.0 Check for the edge case where an identity creates an identity, but they are the same id
        if "created_by_ref" in stix_dict and stix_dict["type"] == "identity":
            if stix_dict["created_by_ref"] == stix_dict["id"]:
                del stix_dict["created_by_ref"]
    except Exception as e:
        logger.exception(e)

    return stix_dict


def make_sro(res, import_type: ImportType):
    """
        High-level function to convert intermediate format into a Stix relationship object
    Args:
        res (): intermediate format
        import_type (): the type of import "STIX21" or "ATT&CK"

    Returns:
        stix_dict {}: a dict containing the stix object
    """
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    stix_dict = {}
    # 2.A) get the typeql properties and relations
    sro_tql_name = res["type"]
    sro_type = ""
    props = res["has"]
    relns = res["relns"]
    sro_sub_rel = ""
    if sro_tql_name in auth["tql_types"]["relations_sro_roles"]:
        sro_sub_rel = sro_tql_name
        sro_type = "relationship"
    elif sro_tql_name == "relationship":
        sro_type = "relationship"
        for has in props:
            if has["typeql"] == "relationship-type":
                sro_sub_rel = has["value"]
                logger.debug(f'found relationship type -> {sro_sub_rel}\n')
                break
    else:
        sro_type = "sighting"
    #
    # Note, Issue, cannot yet tell what to do with a procedure
    #
    attack_object = False
    uses_relation = False
    if sro_tql_name == "procedire":
        sro_sub_rel = sro_tql_name
        is_procedure = True
    else:
        is_procedure = False
    for prop in props:
        if prop["typeql"] == "x-mitre-version":
            attack_object = True

    obj_tql, sro_tql_name, is_list, protocol = sro_type_to_tql(sro_type, sro_sub_rel, import_type, attack_object, uses_relation, is_procedure)

    logger.debug(f'make sro obj_tql ->{obj_tql}\n sro tql name ->{sro_tql_name}')
    # 2.A) get the typeql properties and relations
    props = res["has"]
    relns = res["relns"]
    edges = res["edges"]
    # 2.) setup the match statements first, depending on whether the object is a sighting or a relationship
    # A. If it is a Relationship then find the source and target roles for the relation, and match them in
    if sro_tql_name in auth["reln_name"]["standard_relations"]:
        for stix_rel in auth["reln"]["standard_relations"]:
            if stix_rel["stix"] == sro_tql_name:
                source_role = stix_rel["source"]
                target_role = stix_rel["target"]
                break

        for edge in edges:
            players = edge["player"]
            if edge["role"] == source_role:
                for p in players:
                    stix_dict["source_ref"] = p["stix_id"]
                    break

            elif edge["role"] == target_role:
                for p in players:
                    stix_dict["target_ref"] = p["stix_id"]
                    break
            else:
                logger.error(f'edge role {edge["role"]} not supported')
                return ''

    # B. If it is a Sighting then match the object to the sighting
    elif sro_tql_name == 'sighting':
        is_list = copy.deepcopy(auth["is_lists"]["sro"]["sro"]) + copy.deepcopy(auth["is_lists"]["sro"]["sighting"])
        for edge in edges:
            players = edge["player"]
            if edge["role"] == "sighting-of":
                for p in players:
                    stix_dict["sighting_of_ref"] = p["stix_id"]

            elif edge["role"] == "where-sighted":
                for p in players:
                    if "where_sighted_refs" in stix_dict:
                        stix_dict["where_sighted_refs"].append(p["stix_id"])
                    else:
                        stix_dict["where_sighted_refs"] = []
                        stix_dict["where_sighted_refs"].append(p["stix_id"])
            elif edge["role"] == "observed":
                for p in players:
                    if "observed_data_refs" in stix_dict:
                        stix_dict["observed_data_refs"].append(p["stix_id"])
                    else:
                        stix_dict["observed_data_refs"] = []
                        stix_dict["observed_data_refs"].append(p["stix_id"])
            else:
                logger.error(f'edge role {edge["role"]} not supported')
                return ''

    else:
        logger.error(f'relationship type {sro_tql_name} not supported')
        return ''

    # 3.A) add the properties onto the the object
    stix_dict = make_properties(props, obj_tql, stix_dict, is_list)
    # 3.B) add the relations onto the object
    stix_dict = make_relations(relns, obj_tql, stix_dict, is_list, sro_tql_name, import_type)
    return stix_dict


def make_sco(res: dict, import_type: ImportType):
    """
        High-level function to convert intermediate format into a Stix cyber observable object
    Args:
        res (): intermediate format
        import_type (): the type of import "STIX21" or "ATT&CK"

    Returns:
        stix_dict {}: a dict containing the stix object
    """
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    # - work out the type of object
    stix_dict = {}
    obj_type = res["T_name"]
    # - get the object-specific typeql names, sighting or relationship
    # - work out the type of object
    sco_tql_name = obj_type
    # - get the object-specific typeql names, sighting or relationship
    obj_tql, sco_tql_name, is_list, protocol = sco__type_to_tql(sco_tql_name, import_type)

    # 2.A) get the typeql properties and relations
    props = res["has"]
    relns = res["relns"]

    is_list = copy.deepcopy(auth["is_lists"]["sco"]["sco"]) + copy.deepcopy(auth["is_lists"]["sco"][obj_type])
    # 3.A) add the properties onto the the object
    stix_dict = make_properties(props, obj_tql, stix_dict, is_list)
    # 3.B) add the relations onto the object
    stix_dict = make_relations(relns, obj_tql, stix_dict, is_list, obj_type, import_type)
    return stix_dict


colours_dict = {
    "tlp-amber": {"type": "marking-definition", "spec_version": "2.1",
                  "id": "marking-definition--f88d31f6-486f-44da-b317-01333bde0b82",
                  "created": "2017-01-20T00:00:00.000Z", "definition_type": "tlp",
                  "name": "TLP:AMBER", "definition": {"tlp": "amber"}},
    "tlp-green": {"type": "marking-definition", "spec_version": "2.1",
                  "id": "marking-definition--34098fce-860f-48ae-8e50-ebd3cc5e41da",
                  "created": "2017-01-20T00:00:00.000Z", "definition_type": "tlp",
                  "name": "TLP:GREEN", "definition": {"tlp": "green"}},
    "tlp-white": {"type": "marking-definition", "spec_version": "2.1",
                  "id": "marking-definition--613f2e26-407d-48c7-9eca-b8e91df99dc9",
                  "created": "2017-01-20T00:00:00.000Z", "definition_type": "tlp",
                  "name": "TLP:WHITE", "definition": {"tlp": "white"}},
    "tlp-red": {"type": "marking-definition", "spec_version": "2.1",
                "id": "marking-definition--5e57c739-391a-4eb3-b6be-7d15ca92d5ed",
                "created": "2017-01-20T00:00:00.000Z", "definition_type": "tlp",
                "name": "TLP:RED", "definition": {"tlp": "red"}}
}


def make_meta(res, import_type: ImportType):
    """
        High-level function to convert intermediate format into a Stix meta object
    Args:
        res (): intermediate format
        import_type (): the type of import "STIX21" or "ATT&CK"

    Returns:
        stix_dict {}: a dict containing the stix object
    """
    stix_dict = {}

    obj_type = res["T_name"]
    props = res["has"]
    relns = res["relns"]
    attack_object = False
    if obj_type == "tlp-white" or obj_type == "tlp-green" or obj_type == "tlp-amber" or obj_type == "tlp-red":
        return colours_dict[obj_type]
    j=0
    tmp_val = ""
    for i, prop in enumerate(props):
        if prop["typeql"] == "x-mitre-attack-spec-version":
            attack_object = True
        elif prop["typeql"] == "statement":
            tmp_val = prop["value"]
            j = i

    if tmp_val != "":
        del props[j]

    obj_tql, sdo_tql_name, is_list, protocol = meta_type_to_tql(obj_type, import_type, attack_object)

    # Add the properties onto the the object
    stix_dict = make_properties(props, obj_tql, stix_dict, is_list)
    # Add the relations onto the object
    stix_dict = make_relations(relns, obj_tql, stix_dict, is_list, obj_type, import_type)

    if tmp_val != "":
        stix_dict["definition"] = {"statement": tmp_val}
        stix_dict["definition_type"] = "statement"

    return stix_dict


def make_properties(props, obj_tql, stix_dict, is_list: bool):
    """
        Unpack properties for a stix object (i.e. values at the 'has' level)
    Args:
        props (): a list of proeprties
        obj_tql (): the tql that describes the object
        stix_dict (): the stix dict that we are building
        is_list (): a list of the properties that are lists

    Returns:
        stix_dict {}: a dict containing the stix object
    """
    for prop in props:
        prop_tql_name = prop["typeql"]
        prop_value = prop["value"]
        for stix_name, stix_value in obj_tql.items():
            if prop_tql_name == stix_value:
                # if property is a list, then
                if stix_name in is_list:
                    if stix_name not in stix_dict:
                        stix_dict[stix_name] = []
                        stix_dict[stix_name].append(prop_value)
                        break
                    else:
                        stix_dict[stix_name].append(prop_value)
                        break
                # else property is a value, not a list
                else:
                    stix_dict[stix_name] = prop_value

    return stix_dict


def make_relations(relns, obj_tql, stix_dict, is_list, obj_name, import_type: ImportType):
    """
        Overall branching function for processing all the sub objects
    Args:
        relns (): the list of relations to process
        obj_tql (): the object tql for this stix object
        stix_dict (): the stix dict we are building
        is_list (): the list of properties that are a list
        obj_name (): the stix object these sub objects belong to

    Returns:
        stix_dict {}: a dict containing the stix object
    """
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    for reln in relns:
        reln_name = reln["T_name"]
        if reln_name in auth["tql_types"]["embedded_relations"]:
            stix_dict = make_embedded_relations(reln, reln_name, stix_dict, is_list, obj_name, import_type)

        elif reln_name in auth["tql_types"]["standard_relations"] or reln_name == "sighting":
            stix_dict = make_standard_relations(reln, reln_name, stix_dict, is_list, obj_name, import_type)

        elif reln_name in auth["tql_types"]["key_value_relations"]:
            stix_dict = make_key_value_relations(reln, reln_name, stix_dict, is_list, obj_name, import_type)

        elif reln_name in auth["tql_types"]["list_of_objects"]:
            stix_dict = make_list_of_objects(reln, reln_name, stix_dict, is_list, obj_name, import_type)

        elif reln_name == "v3-extensions" or reln_name == "optional-headers":
            stix_dict = make_object(reln, reln_name, stix_dict, is_list, obj_name, import_type)

        elif reln_name in auth["tql_types"]["extension_relations"]:
            stix_dict = make_extension_relations(reln, reln_name, stix_dict, is_list, obj_name, import_type)

        elif reln_name == "granular-marking":
            stix_dict = make_granular_marking(reln, reln_name, stix_dict, is_list, obj_tql, obj_name)

        elif reln_name == "hashes" or reln_name == "file-header-hashes":
            stix_dict = make_hashes(reln, reln_name, stix_dict)

        else:
            logger.error(f'Error, relation name is {reln_name}')
            break

    return stix_dict


def make_embedded_relations(reln, reln_name, stix_dict, is_list: bool, obj_name, import_type: ImportType):
    """
        Setup embedded relations based on stix-id's
    Args:
        reln (): relation object
        reln_name (): relation name
        stix_dict (): stix dict being built
        is_list (): list of parameters that are lists
        obj_name (): stix object that owns the embedded relation

    Returns:
        stix_dict {}: a dict containing the stix object
    """
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    stix_object_type = obj_name
    for embedded_r in auth["reln"]["embedded_relations"]:
        if reln_name == embedded_r["typeql"]:
            role_pointed = embedded_r["pointed-to"]
            stix_name = embedded_r["rel"]
            role_owner = embedded_r["owner"]

    roles = reln["roles"]
    for role in roles:
        if role["role"] == role_pointed:
            pointed = role
        elif role["role"] == role_owner:
            owner = role
        else:
            logger.error(f'unsupported role in embedded relation {role["role"]}')

    # 1. Is Owner correct, basically my super object?
    # - should be only one object in the owner role list, and its type is the same as my super object type
    own_players = owner["player"]
    own_player = own_players[0]
    if own_player['tql'] == stix_object_type:
        # 2. If 1 is yes, then we want to find the emebedded relation, else not
        pointed_players = pointed["player"]
        for p in pointed_players:
            prop_value = p['stix_id']
            # if property is a list, then
            if stix_name in is_list:
                if stix_name not in stix_dict:
                    stix_dict[stix_name] = []
                    stix_dict[stix_name].append(prop_value)

                else:
                    stix_dict[stix_name].append(prop_value)

            # else property is a value, not a list
            else:
                stix_dict[stix_name] = prop_value
    else:
        return stix_dict

    return stix_dict


def make_standard_relations(reln, reln_name, stix_dict, is_list, obj_name, import_type: ImportType):
    #logger.warning(" make standard relations visited, but not implemented")
    return stix_dict


def make_key_value_relations(reln, reln_name, stix_dict, is_list, obj_type, import_type: ImportType):
    """
        Setup embedded relations based on stix-id's
    Args:
        reln (): relation object
        reln_name (): relation name
        stix_dict (): stix dict being built
        is_list (): list of parameters that are lists
        obj_type (): stix object that owns the embedded relation

    Returns:
        stix_dict {}: a dict containing the stix object
    """
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    for kv_obj in auth["reln"]["key_value_relations"]:
        if reln_name == kv_obj["typeql"]:
            role_pointed = kv_obj["pointed_to"]
            reln_owner = kv_obj["owner"]
            key_name = kv_obj["key"]
            val_name = kv_obj["value"]
            stix_field_name = kv_obj["name"]
            break

    roles = reln["roles"]
    dict_of_kv = {}
    for role in roles:
        if role["role"] == role_pointed:
            players = role["player"]
            for p in players:
                key_value = p["value"]
                props = p['props']
                prop_list = []
                if len(props) > 1:
                    for prop in props:
                        val_value = prop["value"]
                        prop_list.append(val_value)

                    dict_of_kv[key_value] = prop_list

                else:
                    val_value = props[0]["value"]
                    dict_of_kv[key_value] = val_value

    stix_dict[stix_field_name] = dict_of_kv
    return stix_dict


def make_extension_relations(reln, reln_name, stix_dict, is_list, obj_type, import_type: ImportType):
    """
        Setup extension relations based on stix-id's
    Args:
        reln (): relation object
        reln_name (): relation name
        stix_dict (): stix dict being built
        is_list (): list of parameters that are lists
        obj_type (): stix object that owns the embedded relation

    Returns:
        stix_dict {}: a dict containing the stix object
    """
    logger.debug("make extension relations visited")

    local_dict = {}
    local_dict = make_object(reln, reln_name, local_dict, is_list, obj_type, import_type)
    stix_dict["extensions"] = local_dict
    return stix_dict


def make_object(reln, reln_name, stix_dict, is_list, obj_type, import_type: ImportType):
    """
        Setup a sub object
    Args:
        reln (): relation object
        reln_name (): relation name
        stix_dict (): stix dict being built
        is_list (): list of parameters that are lists
        obj_type (): stix object that owns the embedded relation

    Returns:
        stix_dict {}: a dict containing the stix object
    """
    logger.debug("make object visited")
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    for ext_obj in auth["reln"]["extension_relations"]:
        if reln_name == ext_obj["relation"]:
            role_pointed = ext_obj["pointed-to"]
            role_owner = ext_obj["owner"]
            ext_object = ext_obj["object"]
            stix_ext_name = ext_obj["stix"]
            obj_is_list = copy.deepcopy(auth["is_lists"]["sub"][ext_object])
            break

    if ext_object in auth["sub_objects"]:
        obj_props_tql = copy.deepcopy(auth["sub_objects"][ext_object])
    else:
        raise ValueError("no sub-object available")
    roles = reln["roles"]
    ext_data_object = {}
    for role in roles:
        if role["role"] == role_pointed:
            players = role["player"]
            for p in players:
                player = {}
                # get properties for the sub object
                props = p['has']
                for prop in props:
                    prop_name = prop["typeql"]
                    prop_stix_name = prop_value = None
                    for stix, tql in obj_props_tql.items():
                        if prop_name == tql:
                            prop_stix_name = stix
                            prop_value = prop["value"]
                            break
                    # if property is a list, then
                    if prop_stix_name in obj_is_list:
                        if prop_stix_name not in player:
                            player[prop_stix_name] = []
                            player[prop_stix_name].append(prop_value)

                        else:
                            player[prop_stix_name].append(prop_value)

                    # else property is a value, not a list
                    else:
                        player[prop_stix_name] = prop_value
                # now look to see if there are relations
                obj_relns = [k for k, v in obj_props_tql.items() if v == ""]
                sub_relns = p['relns']
                obj_tql = copy.deepcopy(auth["sub_objects"][ext_object])
                new_dict = {}
                new_dict = make_relations(sub_relns, obj_tql, new_dict, is_list, ext_object, import_type)
                for k, v in new_dict.items():
                    player[k] = v

    stix_dict[stix_ext_name] = player
    return stix_dict


def make_list_of_objects(reln, reln_name, stix_dict, is_list, obj_type, import_type: ImportType):
    """
    Args:
        Setup a list of sub objects
        reln (): relation object
        reln_name (): relation name
        stix_dict (): stix dict being built
        is_list (): list of parameters that are lists
        obj_type (): stix object that owns the embedded relation

    Returns:
        stix_dict {}: a dict containing the stix object
    """
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    for l_obj in auth["reln"]["list_of_objects"]:
        if reln_name == l_obj["typeql"]:
            role_pointed = l_obj["pointed_to"]
            reln_object = l_obj["object"]
            stix_field_name = l_obj["name"]
            obj_is_list = copy.deepcopy(auth["is_lists"]["sub"][reln_object])
            logger.debug("obj_is_list: {}".format(obj_is_list))
            logger.debug("reln_object: {}".format(reln_object))
            logger.debug("stix_field_name: {}".format(stix_field_name))
            logger.debug("role_pointed: {}".format(role_pointed))
            break

    if reln_object in auth["sub_objects"]:
        obj_props_tql = copy.deepcopy(auth["sub_objects"][reln_object])
    else:
        raise ValueError("no sub-object available")
    roles = reln["roles"]
    list_of_objects = []
    for role in roles:
        if role["role"] == role_pointed:
            players = role["player"]
            for p in players:
                player = {}
                # get properties for the sub object
                props = p['has']
                sub_relns = p['relns']
                for prop in props:
                    prop_name = prop["typeql"]
                    prop_stix_name = prop_value = None
                    for stix, tql in obj_props_tql.items():
                        if prop_name == tql:
                            prop_stix_name = stix
                            prop_value = prop["value"]
                            break
                    # if property is a list, then
                    if prop_stix_name in obj_is_list:
                        if prop_stix_name not in player:
                            player[prop_stix_name] = []
                            player[prop_stix_name].append(prop_value)

                        else:
                            player[prop_stix_name].append(prop_value)

                    # else property is a value, not a list
                    else:
                        player[prop_stix_name] = prop_value
                # now look to see if there are relations
                obj_relns = [k for k, v in obj_props_tql.items() if v == ""]
                logger.debug(f'sub relns -> {sub_relns}')
                for sub_reln in sub_relns:
                    logger.debug(f'\n\nsub reln -> {sub_reln}')
                    # if the relation is embedded
                    if sub_reln["T_name"] in auth["tql_types"]["embedded_relations"]:
                        for inst in auth["reln"]["embedded_relations"]:
                            if inst["typeql"] == sub_reln["T_name"]:
                                obj_reln_name = inst["typeql"]
                                obj_owner = inst["owner"]
                                obj_pointed = inst["pointed-to"]
                                obj_stix_name = inst["rel"]
                                logger.debug(f'obj_reln_name -> {obj_reln_name}')
                                logger.debug(f'obj_owner -> {obj_owner}')
                                logger.debug(f'obj_pointed -> {obj_pointed}')
                                logger.debug(f'obj_stix_name -> {obj_stix_name}')
                                break

                        local_roles = sub_reln["roles"]
                        for l_r in local_roles:
                            # if the owner role  is considered
                            if l_r["role"] == obj_owner:
                                local_players = l_r["player"]
                                # and the existing object used in the list of objects
                                if local_players[0]["tql"] == reln_object:
                                    for l_r2 in local_roles:
                                        if l_r2["role"] == obj_pointed:
                                            players2 = l_r2["player"]
                                            for p2 in players2:
                                                # then we write the result as an embedded relation
                                                answer = p2['stix_id']
                                                if obj_stix_name in obj_is_list:
                                                    if obj_stix_name not in player:
                                                        player[obj_stix_name] = []
                                                        player[obj_stix_name].append(answer)

                                                    else:
                                                        player[obj_stix_name].append(answer)

                                                else:
                                                    player[obj_stix_name] = answer

                    else:
                        logger.debug(f'unsupported relation for list of objects {sub_reln}')
                        #logger.info(f'embedded --> {stix_models["embedded_relations_typeql"]}')

                list_of_objects.append(player)

    stix_dict[stix_field_name] = list_of_objects
    return stix_dict


def make_granular_marking(reln, reln_name, stix_dict, is_list, obj_tql, obj_type=None):
    """
        Setup granular marking sub object
    Args:
        reln (): relation object
        reln_name (): relation name
        stix_dict (): stix dict being built
        is_list (): list of parameters that are lists
        obj_type (): stix object that owns the embedded relation

    Returns:
        stix_dict {}: a dict containing the stix object
    """
    stix_label = "granular_markings"
    local_marking = {}
    roles = reln["roles"]
    lang_marking = stix_marking = None
    for role in roles:
        if role["role"] == "marking":
            local_p = role["player"][0]
            local_id = local_p["stix_id"]
            local_type = local_id.split('--')[0]
            if local_type == "marking-definition":
                stix_marking = local_id
            else:
                lang_marking = local_id

        elif role["role"] == "marked":
            local_p = role["player"]
            selectors = []
            for p in local_p:
                tql_name = p["tql"]
                stix_name = ''
                for key, value in obj_tql.items():
                    if value == tql_name:
                        stix_name = key
                        break

                # if property being marked is a list, then
                if stix_name in is_list:
                    # find the item in the property list with the same value, as my typeql value
                    tql_value = p["value"]
                    list_to_check = stix_dict[stix_name]
                    for i, item in enumerate(list_to_check):
                        if item == tql_value:
                            stix_name = stix_name + '.[' + str(i) + ']'
                            selectors.append(stix_name)

                            # else property is a value, not a list
                else:
                    selectors.append(stix_name)

    if stix_marking is not None:
        local_marking["marking_ref"] = stix_marking
        local_marking["selectors"] = selectors
    else:
        local_marking["lang"] = lang_marking
        local_marking["selectors"] = selectors

    if "granular_markings" in stix_dict:
        stix_dict["granular_markings"].append(local_marking)
    else:
        stix_dict["granular_markings"] = []
        stix_dict["granular_markings"].append(local_marking)

    return stix_dict


def make_hashes(reln, reln_name, stix_dict):
    """
        Setup hashes sub object
    Args:
        reln (): relation object
        reln_name (): relation name
        stix_dict (): stix dict being built

    Returns:
        stix_dict {}: a dict containing the stix object
    """
    stix_label = "hashes"
    hashes = {}
    roles = reln["roles"]
    for r in roles:
        if r["role"] == "owner":
            own_players = r["player"]
        elif r['role'] == "pointed-to":
            own_players = r["player"]
            for p in own_players:
                hash_type = p["tql"]
                hash_value = p["hash_value"]
                hashes[hash_type] = hash_value

        else:
            logger.error(f" make hashes relation not implemented {r['role']}")

    stix_dict[stix_label] = hashes
    return stix_dict



